Package: mdbr
Title: Work with Microsoft Access Files
Version: 0.3.1
Authors@R: c(
    person("Kiernan", "Nicholls", email = "k5cents@gmail.com", role = c("aut", "cre", "cph"),
           comment = c(ORCID = "0000-0002-9229-7897")),
    person("Bruno", "Tremblay", email = "cran@neoxone.com", role = c("ctb"),
           comment = "DBI interface and bundled mdbtools source"))
Description: Work with Microsoft Access '.mdb' and '.accdb' files using the
    open source 'MDB Tools' library <https://github.com/mdbtools/mdbtools/>.
    The library is compiled and bundled with the package, so no external
    installation is required. Provides high-level helpers for reading tables,
    exporting to CSV or JSON, inspecting table definitions, and running SQL queries.
    Also exposes a full read-only 'DBI' interface for use with standard
    database workflows.
License: GPL-3 | LGPL-2
URL: https://k5cents.github.io/mdbr/, https://github.com/k5cents/mdbr,
        https://github.com/mdbtools/mdbtools/
BugReports: https://github.com/k5cents/mdbr/issues
Imports: DBI, lifecycle, methods, tibble, utils, jsonlite
Suggests: readr, testthat (>= 3.0.0)
Encoding: UTF-8
NeedsCompilation: yes
SystemRequirements: GNU make
Config/testthat/edition: 3
Config/roxygen2/version: 8.0.0
Packaged: 2026-05-24 00:35:50 UTC; kiernan
Author: Kiernan Nicholls [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0002-9229-7897>),
  Bruno Tremblay [ctb] (DBI interface and bundled mdbtools source)
Maintainer: Kiernan Nicholls <k5cents@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-28 11:00:08 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-05-28 13:27:03 UTC; unix
Archs: mdbr.so.dSYM

Package: mdbr
Title: Work with Microsoft Access Files
Version: 0.3.2
Authors@R: c(
    person("Kiernan", "Nicholls", email = "k5cents@gmail.com", role = c("aut", "cre", "cph"),
           comment = c(ORCID = "0000-0002-9229-7897")),
    person("Bruno", "Tremblay", email = "cran@neoxone.com", role = c("ctb"),
           comment = "DBI interface and bundled mdbtools source"))
Description: Work with Microsoft Access '.mdb' and '.accdb' files using the
    open source 'MDB Tools' library <https://github.com/mdbtools/mdbtools/>.
    The library is compiled and bundled with the package, so no external
    installation is required. Provides high-level helpers for reading tables,
    exporting to CSV or JSON, inspecting table definitions, and running SQL queries.
    Also exposes a full read-only 'DBI' interface for use with standard
    database workflows.
License: GPL-3 | LGPL-2
URL: https://k5cents.github.io/mdbr/, https://github.com/k5cents/mdbr,
        https://github.com/mdbtools/mdbtools/
BugReports: https://github.com/k5cents/mdbr/issues
Imports: DBI, lifecycle, methods, tibble, utils, jsonlite
Suggests: readr, testthat (>= 3.0.0)
Encoding: UTF-8
NeedsCompilation: yes
SystemRequirements: GNU make
Config/testthat/edition: 3
Config/roxygen2/version: 8.0.0
Packaged: 2026-07-18 16:33:44 UTC; kiernan
Author: Kiernan Nicholls [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0002-9229-7897>),
  Bruno Tremblay [ctb] (DBI interface and bundled mdbtools source)
Maintainer: Kiernan Nicholls <k5cents@gmail.com>
Repository: CRAN
Date/Publication: 2026-07-18 17:00:08 UTC

# Generated by roxygen2: do not edit by hand

S3method(print,mdblist)
export(export_mdb)
export(mdb)
export(mdb_array)
export(mdb_count)
export(mdb_example)
export(mdb_export)
export(mdb_header)
export(mdb_hexdump)
export(mdb_json)
export(mdb_prop)
export(mdb_queries)
export(mdb_schema)
export(mdb_sql)
export(mdb_tables)
export(mdb_ver)
export(read_mdb)
importFrom(DBI,SQL)
importFrom(DBI,dbCanConnect)
importFrom(DBI,dbClearResult)
importFrom(DBI,dbConnect)
importFrom(DBI,dbDisconnect)
importFrom(DBI,dbExecute)
importFrom(DBI,dbExistsTable)
importFrom(DBI,dbFetch)
importFrom(DBI,dbGetQuery)
importFrom(DBI,dbHasCompleted)
importFrom(DBI,dbIsValid)
importFrom(DBI,dbListFields)
importFrom(DBI,dbListObjects)
importFrom(DBI,dbListTables)
importFrom(DBI,dbQuoteIdentifier)
importFrom(DBI,dbReadTable)
importFrom(DBI,dbSendQuery)
importFrom(jsonlite,toJSON)
importFrom(lifecycle,deprecate_warn)
importFrom(methods,new)
importFrom(methods,setClass)
importFrom(methods,setMethod)
importFrom(tibble,as_tibble)
importFrom(utils,capture.output)
useDynLib(mdbr)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# mdbr <img src='man/figures/logo.png' align="right" height="139" />

<!-- badges: start -->

[![Lifecycle:
experimental](https://img.shields.io/badge/lifecycle-experimental-orange.svg)](https://lifecycle.r-lib.org/articles/stages.html#experimental))
[![CRAN
status](https://www.r-pkg.org/badges/version/mdbr)](https://CRAN.R-project.org/package=mdbr)
[![Codecov test
coverage](https://img.shields.io/codecov/c/github/k5cents/mdbr/master.svg)](https://app.codecov.io/gh/k5cents/mdbr?branch=master)
![Downloads](https://cranlogs.r-pkg.org/badges/grand-total/mdbr) [![R
build
status](https://github.com/k5cents/mdbr/workflows/R-CMD-check/badge.svg)](https://github.com/k5cents/mdbr/actions)
<!-- badges: end -->

The goal of mdbr is to easily access the open source [MDB
Tools](https://github.com/mdbtools/mdbtools) written by Brian Bruns. The
MDB Tools C library is now bundled with the package — no external
installation is required. This package reads proprietary Microsoft
Access files directly and returns standard R data frames.

## Installation

You can install the release version of mdbr from
[CRAN](https://cran.r-project.org/package=mdbr).

``` r
install.packages("mdbr")
```

The development version can be installed from
[GitHub](https://github.com/k5cents/mdbr/).

``` r
# install.packages("remotes")
remotes::install_github("k5cents/mdbr")
```

## Example

``` r
library(mdbr)
```

The package comes with a version of the
[nycflights13](https://github.com/tidyverse/nycflights13) relational
database found with `mdb_examples()`.

The tables in a database can be listed with `mdb_tables()`.

``` r
mdb_tables(ex <- mdb_example())
#> [1] "Airlines" "Airports" "Flights"  "Planes"
```

These tables can be exported as a delimited string or file.

``` r
string <- export_mdb(ex, "Airlines", output = TRUE, delim = "|", quote = "'")
cat(string, sep = "\n")
#> carrier|name
#> '9E'|'Endeavor Air Inc.'
#> 'AA'|'American Airlines Inc.'
#> 'AS'|'Alaska Airlines Inc.'
#> 'B6'|'JetBlue Airways'
#> 'DL'|'Delta Air Lines Inc.'
#> 'EV'|'ExpressJet Airlines Inc.'
#> 'F9'|'Frontier Airlines Inc.'
#> 'FL'|'AirTran Airways Corporation'
#> 'HA'|'Hawaiian Airlines Inc.'
#> 'MQ'|'Envoy Air'
#> 'OO'|'SkyWest Airlines Inc.'
#> 'UA'|'United Air Lines Inc.'
#> 'US'|'US Airways Inc.'
#> 'VX'|'Virgin America'
#> 'WN'|'Southwest Airlines Co.'
#> 'YV'|'Mesa Airlines Inc.'
```

Tables are read directly into R as a tibble with automatic type
coercion.

``` r
read_mdb(ex, "Airports")
#> # A tibble: 1,458 × 8
#>    faa   name                             lat    lon   alt    tz dst   tzone              
#>    <chr> <chr>                          <dbl>  <dbl> <int> <int> <chr> <chr>              
#>  1 04G   Lansdowne Airport               41.1  -80.6  1044    -5 A     America/New_York   
#>  2 06A   Moton Field Municipal Airport   32.5  -85.7   264    -6 A     America/Chicago    
#>  3 06C   Schaumburg Regional             42.0  -88.1   801    -6 A     America/Chicago    
#>  4 06N   Randall Airport                 41.4  -74.4   523    -5 A     America/New_York   
#>  5 09J   Jekyll Island Airport           31.1  -81.4    11    -5 A     America/New_York   
#>  6 0A9   Elizabethton Municipal Airport  36.4  -82.2  1593    -5 A     America/New_York   
#>  7 0G6   Williams County Airport         41.5  -84.5   730    -5 A     America/New_York   
#>  8 0G7   Finger Lakes Regional Airport   42.9  -76.8   492    -5 A     America/New_York   
#>  9 0P2   Shoestring Aviation Airfield    39.8  -76.6  1000    -5 U     America/New_York   
#> 10 0S9   Jefferson County Intl           48.1 -123.    108    -8 A     America/Los_Angeles
#> # ℹ 1,448 more rows
```

The DDL for a table can be retrieved with `mdb_schema(mode = "ddl")`.

``` r
mdb_schema(ex, "Airports", mode = "ddl")
#> [Airports]
#> -- That file uses encoding UTF-8
#> 
#> CREATE TABLE [Airports]
#>  (
#>  [faa]           Text (255),
#>  [name]          Text (255),
#>  [lat]           Double,
#>  [lon]           Double,
#>  [alt]           Long Integer,
#>  [tz]            Integer,
#>  [dst]           Text (255),
#>  [tzone]         Text (255)
#> );
```

Column types are returned as a [readr col
spec](https://readr.tidyverse.org/reference/cols.html) (requires the
**readr** package). Use `condense = TRUE` to collapse columns sharing a
type.

``` r
mdb_schema(ex, "Airports")
#> cols(
#>   faa = col_character(),
#>   name = col_character(),
#>   lat = col_double(),
#>   lon = col_double(),
#>   alt = col_integer(),
#>   tz = col_integer(),
#>   dst = col_character(),
#>   tzone = col_character()
#> )
```

<!-- refs: start -->

<!-- refs: end -->

#' Read a table as data frame
#'
#' Reads a table directly from a Microsoft Access database using the bundled
#' mdbtools C library. Column types are inferred from the MDB schema:
#' integer, double, logical, [POSIXct][base::DateTimeClasses] for DateTime
#' columns, and character otherwise.
#'
#' @param file Path to the Microsoft Access file.
#' @param table Name of the table, list with [mdb_tables()].
#' @param col_names Logical; when `FALSE` columns are named `V1`, `V2`, etc.
#' @param col_types `r lifecycle::badge("deprecated")` Ignored. Type coercion
#'   is now handled automatically by the native reader.
#' @param ... `r lifecycle::badge("deprecated")` Ignored. Extra arguments were
#'   previously forwarded to [readr::read_delim()], which is no longer used.
#' @return A [tibble][tibble::tibble].
#' @importFrom lifecycle deprecate_warn
#' @importFrom tibble as_tibble
#' @examples
#' \dontrun{
#' read_mdb(mdb_example(), "Airlines")
#' }
#' @export
read_mdb <- function(file, table, col_names = TRUE, col_types = NULL, ...) {
  if (missing(table)) {
    stop("Must define a table name, list with mdb_tables()", call. = FALSE)
  }
  if (!is.null(col_types)) {
    lifecycle::deprecate_warn(
      when = "0.3.0",
      what = "read_mdb(col_types)",
      details = "Column types are now inferred automatically by the native reader."
    )
  }
  if (...length() > 0) {
    lifecycle::deprecate_warn(
      when = "0.3.0",
      what = "read_mdb(...)",
      details = "Extra arguments were previously forwarded to readr::read_delim(), which is no longer used."
    )
  }
  path <- .mdb_normalize_path(file)
  raw <- .native_read_table(path, as.character(table))
  df <- .as_data_frame(raw)
  df <- .coerce_mdb_data_frame(df, raw)
  if (!isTRUE(col_names)) {
    names(df) <- paste0("V", seq_along(df))
  }
  tibble::as_tibble(df)
}

library(testthat)
library(mdbr)

test_that("tables can be read as data frames", {
  skip_on_cran()
  skip_if_not(is.loaded("mdbr_version"))
  dat <- read_mdb(mdb_example(), "Flights")
  expect_length(dat, 19)
  expect_s3_class(dat, "tbl_df")
  expect_s3_class(dat$time_hour, "POSIXct")
  expect_type(dat$year, "integer")
  expect_type(dat$carrier, "character")
})

test_that("tables can be read in memory", {
  skip_on_cran()
  skip_if_not(is.loaded("mdbr_version"))
  dat <- read_mdb(mdb_example(), "Flights")
  expect_length(dat, 19)
  expect_s3_class(dat, "tbl_df")
  expect_s3_class(dat$time_hour, "POSIXct")
  expect_type(dat$year, "integer")
  expect_type(dat$carrier, "character")
})

test_that("reading errors without table name", {
  skip_on_cran()
  skip_if_not(is.loaded("mdbr_version"))
  expect_error(read_mdb(mdb_example()))
})

#' List Or Retrieve Queries In An MDB Database
#'
#' `mdb-queries` is a utility program distributed with MDB Tools.
#' Without `query`, it lists the names of all saved queries in the database.
#' With `query`, it returns the SQL text of the named query or queries.
#'
#' @param path Path to `.mdb`/`.accdb` file.
#' @param query Character vector of query names. When `NULL` (the default),
#'   the function lists available query names instead of fetching SQL.
#' @param list Logical; when `TRUE` (the default) and `query` is `NULL`,
#'   return the list of query names. Set to `FALSE` to suppress listing and
#'   return `character(0)`.
#' @param newline Logical; when `TRUE` and `as_text = TRUE`, collapse names
#'   with `"\n"` instead of `delimiter`.
#' @param delimiter Character scalar used to collapse query names when
#'   `as_text = TRUE`. Default `" "`.
#' @param as_text Logical; when `TRUE`, collapses the query name list into a
#'   single string using `delimiter` (or `"\n"` if `newline = TRUE`).
#' @param as_list Logical; defaults to `TRUE`. When `TRUE` and `query` is
#'   supplied, wraps the result in a named `mdblist` object.
#'
#' @return When `query` is `NULL`: a character vector of query names (or a
#'   collapsed string when `as_text = TRUE`). When `query` is supplied and
#'   `as_list = TRUE`: a named `mdblist` of SQL text strings, one per query.
#'   With a single query and `as_list = FALSE`: a character scalar.
#' @examples
#' db <- mdbr:::.mdb_example_nwind_path()
#' if (nzchar(db)) {
#'   mdb_queries(db)
#'   mdb_queries(db, "Orders Qry")
#' }
#' @export
mdb_queries <- function(
  path,
  query = NULL,
  list = TRUE,
  newline = FALSE,
  delimiter = " ",
  as_text = FALSE,
  as_list = TRUE
) {
  path <- .mdb_normalize_path(path)
  if (!is.null(query) && nzchar(as.character(query[[1]]))) {
    query <- as.character(query)
    query <- query[nzchar(query)]
    if (!length(query)) {
      return(character(0))
    }

    out <- stats::setNames(
      lapply(query, function(q) .native_get_query_sql(path, q)),
      query
    )
    out <- lapply(out, as.character)

    if (isTRUE(as_list)) {
      return(.as_mdblist(out))
    }

    if (length(out) > 1L) {
      return(out)
    }

    return(out[[1]])
  }
  if (!isTRUE(list)) {
    return(character(0))
  }
  out <- .native_list_queries(path)
  delim <- if (isTRUE(newline)) "\n" else delimiter
  if (isTRUE(as_text)) {
    return(paste(out, collapse = delim))
  }
  out
}

#' SQL Interface To MDB Tools
#'
#' `mdb-sql` is a utility program distributed with MDB Tools.
#' It allows querying of an MDB database using a limited SQL subset language.
#' The supported SQL is intentionally small: single-table queries, no aggregates,
#' and limited `WHERE` support.
#'
#' In addition to single statements, this wrapper accepts `input` files similar
#' to `mdb-sql -i file`, strips `go` batch terminators, and executes the script
#' one statement at a time.
#'
#' @param path Path to `.mdb`/`.accdb` file.
#' @param statement SQL statement text.
#' @param no_header Logical, equivalent to `-H/--no-header` (for text mode).
#' @param no_footer Logical, equivalent to `-F/--no-footer` (for text mode).
#' @param no_pretty_print Logical, equivalent to `-p/--no-pretty-print`.
#' @param delimiter Delimiter equivalent to `-d/--delimiter` in plain text mode.
#' @param input Input file equivalent to `-i/--input`.
#' @param output Output file equivalent to `-o/--output`.
#' @param as_text Logical; when `TRUE`, returns CLI-like text output.
#'
#' @importFrom utils capture.output
#' @return `data.frame` by default, or character scalar in text mode.
#' @examples
#' db <- mdbr:::.mdb_example_nwind_path()
#' if (nzchar(db)) {
#'   mdb_sql(db, "SELECT [ProductID], [ProductName] FROM [Products] LIMIT 3;")
#' }
#' @export
mdb_sql <- function(
  path,
  statement = NULL,
  no_header = FALSE,
  no_footer = FALSE,
  no_pretty_print = FALSE,
  delimiter = "\t",
  input = NULL,
  output = NULL,
  as_text = FALSE
) {
  if (!is.null(input)) {
    statements <- .mdb_read_sql_input(input)
  } else {
    statements <- statement
  }

  if (
    is.null(statements) ||
      !length(statements) ||
      !any(nzchar(trimws(statements)))
  ) {
    stop("Provide `statement` or `input`.", call. = FALSE)
  }

  statements <- as.character(statements)
  statements <- trimws(statements)
  statements <- statements[nzchar(statements)]

  con <- .mdb_connect(path)
  on.exit(DBI::dbDisconnect(con), add = TRUE)

  text_blocks <- character(0)
  last_df <- data.frame()
  for (stmt in statements) {
    df <- DBI::dbGetQuery(con, stmt)
    last_df <- df

    if (!isTRUE(as_text) && is.null(output) && length(statements) == 1L) {
      return(df)
    }

    if (isTRUE(no_pretty_print)) {
      block <- .mdb_delimited_text(
        df,
        delimiter = delimiter,
        row_delimiter = "\n",
        header = !isTRUE(no_header),
        null = "",
        no_quote = TRUE
      )
    } else {
      lines <- capture.output(print(df, row.names = FALSE))
      if (isTRUE(no_header) && length(lines) > 0L) {
        lines <- lines[-1L]
      }
      block <- paste(lines, collapse = "\n")
    }

    if (!isTRUE(no_footer)) {
      block <- paste0(block, "\n", nrow(df), " row(s)")
    }
    text_blocks <- c(text_blocks, block)
  }

  if (!isTRUE(as_text) && is.null(output)) {
    return(last_df)
  }

  text <- paste(text_blocks, collapse = "\n")
  if (!is.null(output)) {
    writeLines(text, con = output, useBytes = TRUE)
  }
  text
}

#' Count Rows In Table
#'
#' `mdb-count` is a utility program distributed with MDB Tools.
#' It outputs the number of rows in a table.
#'
#' @param path Path to `.mdb`/`.accdb` file.
#' @param table Table name.
#' @param where Optional SQL predicate appended to `WHERE`. This is an R-side
#' extension and is not part of the CLI.
#' @param version Logical; when `TRUE`, return mdbtools version (`--version`).
#'
#' @return Integer row count or version string.
#' @examples
#' db <- mdbr:::.mdb_example_nwind_path()
#' if (nzchar(db)) {
#'   mdb_count(db, "Orders")
#' }
#' @export
mdb_count <- function(path, table = NULL, where = NULL, version = FALSE) {
  if (isTRUE(version)) {
    return(mdb_ver())
  }
  if (is.null(table)) {
    stop("`table` is required unless `version = TRUE`.", call. = FALSE)
  }
  table <- as.character(table[[1]])

  if (is.null(where) || !nzchar(trimws(where))) {
    return(as.integer(.native_table_num_rows(.mdb_normalize_path(path), table)))
  }

  # Read a single column to avoid materializing wide rows when only row count is needed.
  where_sql <- if (!is.null(where) && nzchar(trimws(where))) {
    paste(" WHERE", where)
  } else {
    ""
  }
  fields <- tryCatch(
    .native_list_fields(.mdb_normalize_path(path), table),
    error = function(e) character(0)
  )

  if (length(fields) > 0L && nzchar(fields[[1]])) {
    probe_sql <- sprintf(
      "SELECT %s FROM %s%s",
      .mdb_quote_ident(fields[[1]]),
      .mdb_quote_ident(table),
      where_sql
    )
    rows <- mdb_sql(path, probe_sql)
    return(as.integer(nrow(rows)))
  }

  rows <- mdb_sql(
    path,
    sprintf("SELECT * FROM %s%s", .mdb_quote_ident(table), where_sql)
  )
  as.integer(nrow(rows))
}

#' Generate Schema DDL or Column Types
#'
#' `mdb-schema` is a utility program distributed with MDB Tools. It produces
#' DDL (data definition language) output for the given database, which can be
#' passed to another database to recreate the Access table structure. With
#' `mode = "legacy"` (the default), it returns a
#' [readr col spec][readr::cols()] for the table instead.
#'
#' @param path Path to `.mdb`/`.accdb` file.
#' @param table Table name(s). For `mode = "ddl"`, defaults to all user tables.
#'   For `mode = "legacy"`, exactly one table name must be given.
#' @param mode `"legacy"` (default) returns a [readr::cols()] specification,
#'   matching the k5cents/mdbr canonical API. Requires the \pkg{readr} package.
#'   `"ddl"` returns DDL text.
#' @param condense Logical; only used when `mode = "legacy"`. When `TRUE`,
#'   passes the col spec through [readr::cols_condense()]. Default `FALSE`.
#' @param namespace Prefix identifiers with namespace, equivalent to
#' `-N/--namespace`. Only used when `mode = "ddl"`.
#' @param backend Target DDL dialect. Supported values are `access`, `sybase`,
#' `oracle`, `postgres`, `mysql`, and `sqlite`. Only used when `mode = "ddl"`.
#' @param drop_table Issue `DROP TABLE` statements.
#' @param not_null Include `NOT NULL` constraints.
#' @param default_values Include `DEFAULT` values.
#' @param not_empty Include `CHECK <> ''` constraints.
#' @param comments Include `COMMENT ON` statements.
#' @param indexes Export indexes.
#' @param relations Request foreign key constraints. Current library-mode
#' implementation emits a placeholder comment; full FK export is not yet
#' implemented.
#' @param as_list Logical; defaults to `TRUE`. When `TRUE`, returns a named
#' `mdblist` of DDL text entries keyed by table name. Only used when
#' `mode = "ddl"`.
#'
#' @return When `mode = "legacy"`, a [readr::cols()] specification (optionally
#' condensed via [readr::cols_condense()]). Requires the \pkg{readr} package.
#' When `mode = "ddl"` and `as_list = TRUE`, a named `mdblist` of table-level
#' DDL text.
#' @examples
#' db <- mdbr:::.mdb_example_nwind_path()
#' if (nzchar(db)) {
#'   mdb_schema(db, table = "Products")
#'   mdb_schema(db, table = "Products", mode = "ddl")
#' }
#' @export
mdb_schema <- function(
  path,
  table = NULL,
  mode = c("legacy", "ddl"),
  condense = FALSE,
  namespace = NULL,
  backend = c("access", "sybase", "oracle", "postgres", "mysql", "sqlite"),
  drop_table = FALSE,
  not_null = TRUE,
  default_values = FALSE,
  not_empty = FALSE,
  comments = TRUE,
  indexes = TRUE,
  relations = TRUE,
  as_list = TRUE
) {
  mode <- match.arg(mode)
  if (mode == "legacy") {
    if (is.null(table) || length(table) != 1L) {
      stop(
        "mode = \"legacy\" requires exactly one table name.",
        call. = FALSE
      )
    }
    if (!requireNamespace("readr", quietly = TRUE)) {
      stop(
        "Package \"readr\" is needed for mode = \"legacy\". ",
        "Install it with: install.packages(\"readr\")",
        call. = FALSE
      )
    }
    tbl <- as.character(table)[[1]]
    path_norm <- .mdb_normalize_path(path)
    x <- .native_print_schema(
      path = path_norm,
      table = tbl,
      backend = "access",
      namespace = NULL,
      export_options = .mdb_schema_options()
    )
    x <- strsplit(x[[1]], "\n")[[1]]
    x <- grep("^\t", x, value = TRUE)
    x <- gsub("\t{3}", "|", x)
    x <- gsub("^\t", "", x)
    x <- gsub(",\\s*$", "", x)
    x <- gsub("\\[|\\]", "", x)
    x <- gsub("\\s\\(\\d+\\).*", "", x)
    x <- gsub("\\sNOT NULL", "", x)
    y <- matrix(
      data = unlist(strsplit(x, "\\|")),
      ncol = 2,
      byrow = TRUE
    )
    z <- vapply(y[, 2], list_switch, character(1), mdb_col_types)
    names(z) <- y[, 1]
    spec <- readr::as.col_spec(z)
    if (isTRUE(condense)) {
      return(readr::cols_condense(spec))
    }
    return(spec)
  }

  table_vec <- if (is.null(table)) {
    NULL
  } else {
    v <- as.character(table)
    v[nzchar(v)]
  }

  backend <- match.arg(backend)
  export_options <- .mdb_schema_options(
    drop_table = drop_table,
    not_null = not_null,
    default_values = default_values,
    not_empty = not_empty,
    comments = comments,
    indexes = indexes,
    relations = relations
  )
  path <- .mdb_normalize_path(path)
  namespace_val <- if (!is.null(namespace)) {
    as.character(namespace[[1]])
  } else {
    NULL
  }

  # C now returns a named character vector: names = table names, values = DDL.
  # Calling with table = NULL fetches all user tables in a single file open.
  # For a single specific table, pass it directly so C can error if not found.
  if (!is.null(table_vec) && length(table_vec) == 1L) {
    result <- .native_print_schema(
      path = path,
      table = table_vec[[1]],
      backend = backend,
      namespace = namespace_val,
      export_options = export_options
    )
  } else {
    result <- .native_print_schema(
      path = path,
      table = NULL,
      backend = backend,
      namespace = namespace_val,
      export_options = export_options
    )
    if (!is.null(table_vec) && length(table_vec) > 1L) {
      missing_tables <- setdiff(table_vec, names(result))
      if (length(missing_tables) > 0L) {
        stop(
          sprintf(
            "Table(s) not found for schema export: %s",
            paste(missing_tables, collapse = ", ")
          ),
          call. = FALSE
        )
      }
      result <- result[match(table_vec, names(result))]
    }
  }

  if (isTRUE(as_list)) {
    return(.as_mdblist(as.list(result)))
  }

  # as_list = FALSE: single table or all-tables-concatenated returns a scalar
  # string; multiple specific tables returns a plain named list.
  if (length(result) == 1L || is.null(table_vec)) {
    return(paste(result, collapse = "\n"))
  }

  as.list(result)
}

#' Return MDB File Format Or MDB Tools Version
#'
#' `mdb-ver` will return a single line of output corresponding to the program
#' that produced the file: `JET3` (for files produced by Access 97), `JET4`
#' (Access 2000, XP and 2003), `ACE12` (Access 2007), `ACE14` (Access 2010),
#' `ACE15` (Access 2013), or `ACE16` (Access 2016).
#'
#' @param path Optional database path. When omitted, the wrapper returns the
#' mdbtools package version for backward compatibility.
#' @param version Logical, equivalent to `-M/--version`.
#'
#' @return Single character string with file format or mdbtools version.
#' @examples
#' mdb_ver()
#' db <- mdbr:::.mdb_example_nwind_path()
#' if (nzchar(db)) {
#'   mdb_ver(db)
#' }
#' @export
mdb_ver <- function(path = NULL, version = FALSE) {
  if (isTRUE(version) || is.null(path)) {
    return(.native_mdbtools_version())
  }
  .native_file_format(.mdb_normalize_path(path))
}

#' Export Table As List Columns (mdb-array mimic)
#'
#' `mdb-array(1)` emits C source; this wrapper returns a named R list of columns
#' while keeping equivalent database/table inputs.
#'
#' @param path Path to `.mdb`/`.accdb` file.
#' @param table Table name.
#' @param columns Optional character vector of columns.
#' @param n Optional row limit (`LIMIT n`).
#'
#' @return Named list, one entry per selected column.
#' @examples
#' db <- mdbr:::.mdb_example_nwind_path()
#' if (nzchar(db)) {
#'   arr <- mdb_array(db, "Products", columns = c("ProductID", "ProductName"), n = 2)
#'   str(arr)
#' }
#' @export
mdb_array <- function(path, table, columns = NULL, n = -1L) {
  table <- as.character(table[[1]])
  col_sql <- if (is.null(columns) || !length(columns)) {
    "*"
  } else {
    paste(.mdb_quote_ident(columns), collapse = ", ")
  }
  sql <- sprintf("SELECT %s FROM %s", col_sql, .mdb_quote_ident(table))
  if (!is.null(n) && is.finite(n) && n >= 0) {
    sql <- paste(sql, "LIMIT", as.integer(n))
  }
  as.list(mdb_sql(path, sql))
}

#' Export Data In An MDB Table To CSV Or INSERT SQL
#'
#' `mdb-export` is a utility program distributed with MDB Tools.
#' It produces CSV output for the given table. Such output is suitable for
#' importation into databases or spreadsheets.
#'
#' Used with `insert`, it outputs backend-specific SQL `INSERT` statements.
#' Most formatting options also apply in insert mode.
#'
#' @param path Path to `.mdb`/`.accdb` file.
#' @param table Table name.
#' @param no_header Logical, equivalent to `-H/--no-header`.
#' @param delimiter Equivalent to `-d/--delimiter`.
#' @param row_delimiter Equivalent to `-R/--row-delimiter`.
#' @param no_quote Equivalent to `-Q/--no-quote`.
#' @param quote Equivalent to `-q/--quote`.
#' @param escape Equivalent to `-X/--escape`.
#' @param escape_invisible Equivalent to `-e/--escape-invisible`.
#' @param date_format Equivalent to `-D/--date-format`.
#' @param datetime_format Equivalent to `-T/--datetime-format`.
#' @param null Equivalent to `-0/--null`.
#' @param bin Binary mode (`strip`, `raw`, `octal`, `hex`) for parity.
#' @param boolean_words Equivalent to `-B/--boolean-words`.
#' @param insert Backend for `-I/--insert` mode.
#' @param namespace Equivalent to `-N/--namespace`.
#' @param batch_size Equivalent to `-S/--batch-size`.
#' @param n Optional row limit (`LIMIT n`).
#'
#' @return Character scalar containing CSV or SQL INSERT text.
#' @examples
#' db <- mdbr:::.mdb_example_nwind_path()
#' if (nzchar(db)) {
#'   cat(mdb_export(db, "Products", n = 2))
#' }
#' @export
mdb_export <- function(
  path,
  table,
  no_header = FALSE,
  delimiter = ",",
  row_delimiter = "\n",
  no_quote = FALSE,
  quote = '"',
  escape = NULL,
  escape_invisible = FALSE,
  date_format = "%Y-%m-%d",
  datetime_format = "%Y-%m-%d %H:%M:%S",
  null = "",
  bin = c("strip", "raw", "octal", "hex"),
  boolean_words = FALSE,
  insert = NULL,
  namespace = NULL,
  batch_size = 1L,
  n = -1L
) {
  bin <- match.arg(bin)
  if (bin != "strip") {
    warning(
      "`bin` modes other than 'strip' are not fully implemented in library-only mode.",
      call. = FALSE
    )
  }

  df <- .mdb_query_table(path, table, n = n)
  df <- .mdb_apply_datetime_formats(
    df,
    date_format = date_format,
    datetime_format = datetime_format
  )
  df <- .mdb_apply_boolean_words(df, boolean_words = boolean_words)

  if (identical(bin, "strip")) {
    for (nm in names(df)) {
      if (is.raw(df[[nm]])) {
        df[[nm]] <- NA_character_
      }
    }
  }

  if (is.null(insert)) {
    return(.mdb_delimited_text(
      df,
      delimiter = delimiter,
      row_delimiter = row_delimiter,
      header = !isTRUE(no_header),
      null = null,
      no_quote = no_quote,
      quote = quote,
      escape = escape,
      escape_invisible = escape_invisible
    ))
  }

  backend <- match.arg(
    insert,
    c("access", "sybase", "oracle", "postgres", "mysql", "sqlite")
  )
  batch_size <- max(1L, as.integer(batch_size[[1]]))
  target <- .mdb_quote_ident(as.character(table[[1]]))
  if (!is.null(namespace) && nzchar(namespace)) {
    target <- paste0(.mdb_quote_ident(namespace), ".", target)
  }

  cols <- paste(.mdb_quote_ident(names(df)), collapse = ", ")
  out <- character(0)
  if (nrow(df) == 0L) {
    return("")
  }

  for (i in seq(1L, nrow(df), by = batch_size)) {
    j <- min(i + batch_size - 1L, nrow(df))
    chunk <- df[i:j, , drop = FALSE]
    vals <- apply(chunk, 1L, function(row) {
      paste0(
        "(",
        paste(
          vapply(row, .mdb_sql_literal, FUN.VALUE = character(1)),
          collapse = ", "
        ),
        ")"
      )
    })
    stmt <- sprintf(
      "INSERT INTO %s (%s) VALUES %s;",
      target,
      cols,
      paste(vals, collapse = ", ")
    )
    if (backend %in% c("access", "oracle", "sybase") && length(vals) > 1L) {
      stmt <- paste(
        sprintf("INSERT INTO %s (%s) VALUES %s;", target, cols, vals),
        collapse = "\n"
      )
    }
    out <- c(out, stmt)
  }
  paste(out, collapse = "\n")
}

#' MDB Header Summary (mdb-header mimic)
#'
#' `mdb-header(1)` writes C files; this wrapper returns a structured summary.
#'
#' @param path Path to `.mdb`/`.accdb` file.
#'
#' @return Named list with version, table names and query names.
#' @examples
#' db <- mdbr:::.mdb_example_nwind_path()
#' if (nzchar(db)) {
#'   mdb_header(db)
#' }
#' @export
mdb_header <- function(path) {
  list(
    version = mdb_ver(),
    tables = mdb_tables(path, system = TRUE),
    queries = mdb_queries(path)
  )
}

#' Hexdump MDB File (mdb-hexdump mimic)
#'
#' @param path Path to file.
#' @param pagenumber Optional page index (0-based) like `mdb-hexdump file [pagenumber]`.
#' @param page_size Page size in bytes (default 4096 for modern Jet/ACE).
#' @param n Number of bytes to emit.
#'
#' @return Single hexadecimal string.
#' @examples
#' db <- mdbr:::.mdb_example_nwind_path()
#' if (nzchar(db)) {
#'   mdb_hexdump(db, n = 16)
#' }
#' @export
mdb_hexdump <- function(path, pagenumber = NULL, page_size = 4096L, n = 256L) {
  con <- file(path, "rb")
  on.exit(close(con), add = TRUE)
  if (!is.null(pagenumber)) {
    seek(
      con,
      where = as.integer(pagenumber) * as.integer(page_size),
      origin = "start"
    )
  }
  bytes <- readBin(con, what = "raw", n = as.integer(n))
  paste(sprintf("%02X", as.integer(bytes)), collapse = " ")
}

#' Import CSV Into MDB (mdb-import mimic)
#'
#' `mdb-import(1)` writes to MDB files. This package is read-only, so this
#' function validates CLI-like options and then errors.
#'
#' @param path Path to `.mdb`/`.accdb` file.
#' @param table Table name.
#' @param csvfile CSV file path.
#' @param header_lines Equivalent to `-H/--header` skipped lines.
#' @param delimiter Equivalent to `-d/--delimiter`.
#'
#' @return No return; always errors in read-only mode.
#' @keywords internal
mdb_import <- function(
  path,
  table,
  csvfile,
  header_lines = 0L,
  delimiter = ","
) {
  stop(
    "`mdb_import()` is not available: this package currently supports read-only MDB/ACCDB operations.",
    call. = FALSE
  )
}

#' Export Data In An MDB Table To JSON
#'
#' `mdb-json` is a utility program distributed with MDB Tools.
#' It produces JSON output for the given table. Such output is suitable for
#' parsing in a variety of languages.
#'
#' @param path Path to `.mdb`/`.accdb` file.
#' @param table Table name.
#' @param date_format Equivalent to `-D/--date-format`.
#' @param time_format Equivalent to `-T/--time-format`.
#' @param no_unprintable Equivalent to `-U/--no-unprintable`.
#' @param n Optional row limit.
#'
#' @return JSON string.
#' @importFrom jsonlite toJSON
#' @examples
#' db <- mdbr:::.mdb_example_nwind_path()
#' if (nzchar(db) && requireNamespace("jsonlite", quietly = TRUE)) {
#'   mdb_json(db, "Products", n = 2)
#' }
#' @export
mdb_json <- function(
  path,
  table,
  date_format = "%Y-%m-%d",
  time_format = "%Y-%m-%d %H:%M:%S",
  no_unprintable = FALSE,
  n = -1L
) {
  if (!requireNamespace("jsonlite", quietly = TRUE)) {
    stop("`mdb_json()` requires the `jsonlite` package.", call. = FALSE)
  }
  df <- .mdb_query_table(path, table, n = n)
  df <- .mdb_apply_datetime_formats(
    df,
    date_format = date_format,
    datetime_format = time_format
  )
  df <- .mdb_apply_unprintable(df, no_unprintable = no_unprintable)
  jsonlite::toJSON(df, dataframe = "rows", na = "null", auto_unbox = TRUE)
}

#' Get Properties List From MDB Database
#'
#' `mdb-prop` retrieves properties for one or more objects in an MDB database.
#' `name` is the name of the table, query, or other object.
#' `propcol` is the name of the `MSysObjects` column containing properties and
#' defaults to `LvProp`.
#'
#' @param path Path to `.mdb`/`.accdb` file.
#' @param name Object name (`table`, `query`, etc.); may be a character vector.
#' @param propcol Property column name. Defaults to `LvProp`.
#' @param version Logical; when `TRUE`, return mdbtools version.
#' @param as_list Logical; defaults to `TRUE`. When `TRUE`, wraps the result in
#' an `mdblist` object so the dedicated print method is used.
#'
#' @return A named list with one entry per element of `name`. Each entry is
#' itself a named list of named character vectors, one per property block
#' (e.g. `"(none)"` for table-level properties, or a column/field name).
#' Access individual values with `p[["Orders"]][["(none)"]]["Description"]`.
#' @examples
#' db <- mdbr:::.mdb_example_nwind_path()
#' if (nzchar(db)) {
#'   p <- mdb_prop(db, "Orders")
#'   p[["Orders"]][["(none)"]]["Description"]
#'   p2 <- mdb_prop(db, c("Orders", "Orders Qry"))
#' }
#' @export
mdb_prop <- function(
  path,
  name = NULL,
  propcol = "LvProp",
  version = FALSE,
  as_list = TRUE
) {
  if (isTRUE(version)) {
    return(mdb_ver())
  }
  if (is.null(name)) {
    stop("`name` is required unless `version = TRUE`.", call. = FALSE)
  }

  path <- .mdb_normalize_path(path)
  name <- as.character(name)
  name <- name[nzchar(name)]
  if (!length(name)) {
    stop(
      "`name` must contain at least one non-empty object name.",
      call. = FALSE
    )
  }

  propcol <- if (!is.null(propcol) && nzchar(propcol)) {
    as.character(propcol[[1]])
  } else {
    NULL
  }

  out <- stats::setNames(
    lapply(name, function(nm) .native_prop_dump(path, nm, propcol)),
    name
  )

  if (isTRUE(as_list)) .as_mdblist(out) else out
}

library(mdbr)

sample_accdb <- testthat::test_path(
  "mdbtestdata",
  "data",
  "ASampleDatabase.accdb"
)
sample_mdb <- testthat::test_path("mdbtestdata", "data", "nwind.mdb")
sample_sql <- testthat::test_path("mdbtestdata", "sql", "nwind.sql")

read_sql_statements <- function(path) {
  lines <- readLines(path, warn = FALSE, encoding = "UTF-8")
  lines <- trimws(lines)
  lines <- lines[nzchar(lines)]
  lines <- lines[!startsWith(lines, "#")]
  lines
}

test_that("driver constructor returns DBIDriver", {
  drv <- mdb()
  expect_s4_class(drv, "MdbDriver")
})

test_that("native symbols are loaded", {
  expect_true(is.loaded("mdbr_list_tables"))
  expect_true(is.loaded("mdbr_list_queries"))
  expect_true(is.loaded("mdbr_list_fields"))
  expect_true(is.loaded("mdbr_read_table"))
  expect_true(is.loaded("mdbr_run_query"))
  expect_true(is.loaded("mdbr_get_query_sql"))
})

test_that("character dbConnect dispatch works for accdb path", {
  skip_if_not(file.exists(sample_accdb))

  conn <- DBI::dbConnect(sample_accdb)
  on.exit(DBI::dbDisconnect(conn), add = TRUE)

  expect_true(DBI::dbIsValid(conn))
})

test_that("basic DBI methods operate on sample accdb", {
  skip_if_not(file.exists(sample_accdb))

  conn <- DBI::dbConnect(mdb(), dbname = sample_accdb)
  on.exit(DBI::dbDisconnect(conn), add = TRUE)

  tables <- DBI::dbListTables(conn)
  expect_true(length(tables) > 0)
  expect_true("Asset Items" %in% tables)

  target <- "Asset Items"
  expect_true(DBI::dbExistsTable(conn, target))

  fields <- DBI::dbListFields(conn, target)
  expect_true(length(fields) >= 1)

  df <- DBI::dbReadTable(conn, target)
  expect_s3_class(df, "data.frame")
})

test_that("dbReadTable applies type coercion from MDB metadata", {
  skip_if_not(file.exists(sample_mdb))

  conn <- DBI::dbConnect(sample_mdb)
  on.exit(DBI::dbDisconnect(conn), add = TRUE)

  df <- DBI::dbReadTable(conn, "Umsätze")
  expect_s3_class(df, "data.frame")

  expect_type(df$OrderID, "integer")
  expect_type(df$Freight, "double")
  expect_s3_class(df$ShippedDate, "POSIXct")
  expect_identical(attr(df$ShippedDate, "tzone"), "UTC")

  formatted <- format(df$ShippedDate, "%Y-%m-%d %H:%M:%S", tz = "UTC")
  expect_true(all(
    is.na(df$ShippedDate) |
      grepl(
        "^[0-9]{4}-[0-9]{2}-[0-9]{2} [0-9]{2}:[0-9]{2}:[0-9]{2}$",
        formatted
      )
  ))
})

test_that("query roundtrip works", {
  skip_if_not(file.exists(sample_mdb))

  conn <- DBI::dbConnect(sample_mdb)
  on.exit(DBI::dbDisconnect(conn), add = TRUE)

  tables <- DBI::dbListTables(conn)
  expect_true("Umsätze" %in% tables)

  out <- DBI::dbGetQuery(conn, "SELECT * FROM [Umsätze] LIMIT 2;")
  expect_s3_class(out, "data.frame")
  expect_lte(nrow(out), 2)
  expect_true("OrderID" %in% names(out))
})

test_that("dbGetQuery applies type coercion from MDB metadata", {
  skip_if_not(file.exists(sample_mdb))

  conn <- DBI::dbConnect(sample_mdb)
  on.exit(DBI::dbDisconnect(conn), add = TRUE)

  out <- DBI::dbGetQuery(
    conn,
    "SELECT [OrderID], [Freight], [ShippedDate] FROM [Umsätze] LIMIT 5;"
  )

  expect_type(out$OrderID, "integer")
  expect_type(out$Freight, "double")
  expect_s3_class(out$ShippedDate, "POSIXct")
  expect_identical(attr(out$ShippedDate, "tzone"), "UTC")

  formatted <- format(out$ShippedDate, "%Y-%m-%d %H:%M:%S", tz = "UTC")
  expect_true(all(
    is.na(out$ShippedDate) |
      grepl(
        "^[0-9]{4}-[0-9]{2}-[0-9]{2} [0-9]{2}:[0-9]{2}:[0-9]{2}$",
        formatted
      )
  ))
})

test_that("dbGetQuery does not execute saved Access query names yet", {
  skip_if_not(file.exists(sample_mdb))

  conn <- DBI::dbConnect(sample_mdb)
  on.exit(DBI::dbDisconnect(conn), add = TRUE)

  expect_error(
    DBI::dbGetQuery(conn, "SELECT * FROM [Current Product List] LIMIT 3;"),
    regexp = "not a table|Got no result"
  )
})

test_that("mdb_queries lists saved queries and extracts SQL", {
  skip_if_not(file.exists(sample_mdb))

  queries <- mdb_queries(sample_mdb)
  expect_true("Current Product List" %in% queries)

  sql <- mdb_queries(sample_mdb, query = "Current Product List")
  expect_s3_class(sql, "mdblist")
  expect_identical(names(sql), "Current Product List")
  expect_true(grepl(
    "^SELECT",
    sql[["Current Product List"]],
    ignore.case = TRUE
  ))
})

test_that("dbListObjects returns DBI-shaped data frame with tables and queries", {
  skip_if_not(file.exists(sample_mdb))

  conn <- DBI::dbConnect(sample_mdb)
  on.exit(DBI::dbDisconnect(conn), add = TRUE)

  objs <- DBI::dbListObjects(conn)
  expect_s3_class(objs, "data.frame")
  expect_identical(names(objs)[1:2], c("table", "is_prefix"))
  expect_true(is.list(objs$table))
  expect_type(objs$is_prefix, "logical")
  expect_true(all(!objs$is_prefix))

  obj_names <- vapply(objs$table, as.character, FUN.VALUE = character(1))
  expect_true("Products" %in% obj_names)
  expect_true("Current Product List" %in% obj_names)
})

test_that("mdb_ver and mdb_schema work without system CLI", {
  skip_if_not(file.exists(sample_mdb))

  ver <- mdb_ver()
  expect_type(ver, "character")
  expect_gt(nchar(ver), 0)

  file_ver <- mdb_ver(sample_mdb)
  expect_identical(file_ver, "JET3")

  ddl <- mdb_schema(
    sample_mdb,
    table = "Products",
    mode = "ddl",
    backend = "postgres",
    as_list = FALSE
  )
  expect_type(ddl, "character")
  expect_true(grepl("CREATE TABLE", ddl, fixed = TRUE))
})

test_that("mdb_prop returns named list of named lists for single object", {
  skip_if_not(file.exists(sample_mdb))

  props <- mdb_prop(sample_mdb, "Orders")
  expect_type(props, "list")
  expect_identical(names(props), "Orders")
  expect_type(props[["Orders"]], "list")
  expect_gt(length(props[["Orders"]]), 0L)
})

test_that("mdb_prop returns named list of named lists for multiple objects", {
  skip_if_not(file.exists(sample_mdb))

  props <- mdb_prop(sample_mdb, c("Orders", "Orders Qry"))
  expect_type(props, "list")
  expect_true(all(c("Orders", "Orders Qry") %in% names(props)))
  expect_type(props[["Orders"]], "list")
  expect_gt(length(props[["Orders"]]), 0L)
})

test_that("mdb_queries can return named mdblist for multiple query SQL texts", {
  skip_if_not(file.exists(sample_mdb))

  queries <- mdb_queries(sample_mdb)
  supported <- queries[vapply(
    queries,
    function(q) {
      !inherits(
        try(mdb_queries(sample_mdb, query = q, as_list = FALSE), silent = TRUE),
        "try-error"
      )
    },
    logical(1)
  )]
  target <- head(supported, 2)
  skip_if_not(length(target) >= 1)

  sqls <- mdb_queries(sample_mdb, query = target, as_list = TRUE)
  expect_s3_class(sqls, "mdblist")
  expect_true(all(target %in% names(sqls)))
  expect_true(all(vapply(
    sqls,
    function(x) is.character(x) && nchar(x) > 0,
    logical(1)
  )))
})

test_that("mdb_queries query SQL output is mdblist by default", {
  skip_if_not(file.exists(sample_mdb))

  queries <- mdb_queries(sample_mdb)
  supported <- queries[vapply(
    queries,
    function(q) {
      !inherits(
        try(mdb_queries(sample_mdb, query = q, as_list = FALSE), silent = TRUE),
        "try-error"
      )
    },
    logical(1)
  )]
  target <- head(supported, 2)
  skip_if_not(length(target) >= 1)

  sqls <- mdb_queries(sample_mdb, query = target)
  expect_s3_class(sqls, "mdblist")
  expect_true(all(target %in% names(sqls)))
})

test_that("mdb_queries mirrors CLI placeholder for unsupported saved-query layout", {
  skip_if_not(file.exists(sample_mdb))

  q <- "Summary of Sales by Quarter"
  skip_if_not(q %in% mdb_queries(sample_mdb))

  sql <- mdb_queries(sample_mdb, query = q, as_list = FALSE)
  expect_identical(sql, "SELECT  FROM  ")
})

test_that("mdb_schema selected table output is mdblist by default", {
  skip_if_not(file.exists(sample_mdb))

  ddl <- mdb_schema(sample_mdb, table = "Products", mode = "ddl")
  expect_s3_class(ddl, "mdblist")
  expect_identical(names(ddl), "Products")
  expect_true(grepl("CREATE TABLE", ddl[["Products"]], fixed = TRUE))
})

test_that("mdb_schema can return named mdblist for selected tables", {
  skip_if_not(file.exists(sample_mdb))

  ddl <- mdb_schema(sample_mdb, table = c("Products", "Orders"), mode = "ddl", as_list = TRUE)
  expect_s3_class(ddl, "mdblist")
  expect_true(all(c("Products", "Orders") %in% names(ddl)))
  expect_true(all(vapply(
    ddl,
    function(x) grepl("CREATE TABLE", x, fixed = TRUE),
    logical(1)
  )))
})

test_that("mdb_schema with no table returns mdblist by default", {
  skip_if_not(file.exists(sample_mdb))

  ddl <- mdb_schema(sample_mdb, mode = "ddl")
  expect_s3_class(ddl, "mdblist")
  expect_true(length(ddl) > 0)
  expect_true("Products" %in% names(ddl))
  expect_true(grepl("CREATE TABLE", ddl[[1]], fixed = TRUE))
})

test_that("mdb_schema output does not include legacy banner", {
  skip_if_not(file.exists(sample_mdb))

  ddl <- mdb_schema(sample_mdb, table = "Products", mode = "ddl", as_list = FALSE)
  expect_false(grepl(
    "MDB Tools - A library for reading MS Access database files",
    ddl,
    fixed = TRUE
  ))
})

test_that("mdb-tables and mdb-queries option mimics behave", {
  skip_if_not(file.exists(sample_mdb))

  table_vec <- mdb_tables(sample_mdb)
  expect_type(table_vec, "character")
  expect_gte(length(table_vec), 1L)

  query_names <- mdb_queries(sample_mdb, list = TRUE)
  expect_type(query_names, "character")

  query_text <- mdb_queries(
    sample_mdb,
    list = TRUE,
    newline = TRUE,
    as_text = TRUE
  )
  expect_type(query_text, "character")

  typed_any <- mdb_tables(sample_mdb, type = "any", show_type = TRUE)
  expect_true(any(grepl("^query ", typed_any)))
})

test_that("mdb-export and mdb-sql option mimics return text output", {
  skip_if_not(file.exists(sample_mdb))

  sql_text <- mdb_sql(
    sample_mdb,
    "SELECT [OrderID], [Freight] FROM [Umsätze] LIMIT 2;",
    as_text = TRUE,
    no_pretty_print = TRUE,
    no_footer = TRUE,
    delimiter = "|"
  )
  expect_type(sql_text, "character")
  expect_true(grepl("\\|", sql_text) || nzchar(sql_text))

  export_text <- mdb_export(
    sample_mdb,
    "Umsätze",
    no_header = TRUE,
    delimiter = ";",
    row_delimiter = "\n",
    no_quote = TRUE,
    n = 2
  )
  expect_type(export_text, "character")
  expect_true(nchar(export_text) > 0)

  # Categories includes binary-like content in nwind; export should not error
  # due to locale-invalid bytes during quoting.
  expect_type(mdb_export(sample_mdb, "Categories", n = 1), "character")
})

test_that("mdb_count fallback preserves WHERE and matches projected-row counts", {
  skip_if_not(file.exists(sample_mdb))

  total <- mdb_count(sample_mdb, "Umsätze")
  filtered <- mdb_count(sample_mdb, "Umsätze", where = "[OrderID] > 11000")

  projected_total <- nrow(mdb_sql(
    sample_mdb,
    "SELECT [OrderID] FROM [Umsätze]"
  ))
  projected_filtered <- nrow(mdb_sql(
    sample_mdb,
    "SELECT [OrderID] FROM [Umsätze] WHERE [OrderID] > 11000"
  ))

  expect_identical(total, as.integer(projected_total))
  expect_identical(filtered, as.integer(projected_filtered))
  expect_lt(filtered, total)
})

test_that("mdb_count without WHERE uses metadata row count semantics", {
  skip_if_not(file.exists(sample_mdb))

  count <- mdb_count(sample_mdb, "Umsätze")
  conn <- DBI::dbConnect(sample_mdb)
  on.exit(DBI::dbDisconnect(conn), add = TRUE)
  table_rows <- DBI::dbReadTable(conn, "Umsätze")
  expect_identical(count, as.integer(nrow(table_rows)))
})

test_that("test_script.sh command set is covered by mimic wrappers", {
  skip_if_not(file.exists(sample_accdb))
  skip_if_not(file.exists(sample_mdb))

  json_accdb <- mdb_json(sample_accdb, "Asset Items", n = 3)
  expect_type(json_accdb, "character")
  expect_true(grepl("\\[|\\{", json_accdb))

  json_mdb <- mdb_json(sample_mdb, "Umsätze", n = 3)
  expect_type(json_mdb, "character")
  expect_true(grepl("\\[|\\{", json_mdb))

  count_accdb <- mdb_count(sample_accdb, "Asset Items")
  expect_type(count_accdb, "integer")
  expect_gte(count_accdb, 0L)

  count_mdb <- mdb_count(sample_mdb, "Umsätze")
  expect_type(count_mdb, "integer")
  expect_gte(count_mdb, 0L)

  prop_accdb <- mdb_prop(sample_accdb, name = "Asset Items")
  expect_type(prop_accdb, "list")
  expect_type(prop_accdb[["Asset Items"]], "list")
  expect_gt(length(prop_accdb[["Asset Items"]]), 0L)

  prop_mdb <- mdb_prop(sample_mdb, name = "Umsätze")
  expect_type(prop_mdb, "list")
  expect_type(prop_mdb[["Umsätze"]], "list")
  expect_gt(length(prop_mdb[["Umsätze"]]), 0L)

  schema_accdb <- mdb_schema(sample_accdb, mode = "ddl", as_list = FALSE)
  expect_type(schema_accdb, "character")
  expect_true(grepl("CREATE TABLE", schema_accdb, fixed = TRUE))

  schema_mdb <- mdb_schema(sample_mdb, mode = "ddl", as_list = FALSE)
  expect_type(schema_mdb, "character")
  expect_true(grepl("CREATE TABLE", schema_mdb, fixed = TRUE))

  tables_accdb <- mdb_tables(sample_accdb)
  expect_type(tables_accdb, "character")
  expect_true("Asset Items" %in% tables_accdb)

  tables_mdb <- mdb_tables(sample_mdb)
  expect_type(tables_mdb, "character")
  expect_true("Umsätze" %in% tables_mdb)

  ver_accdb <- mdb_ver(sample_accdb)
  ver_mdb <- mdb_ver(sample_mdb)
  expect_true(ver_accdb %in% c("JET4", "ACE12", "ACE14", "ACE15", "ACE16"))
  expect_identical(ver_mdb, "JET3")

  queries_accdb <- mdb_queries(sample_accdb)
  expect_type(queries_accdb, "character")
  if (!"qryCostsSummedByOwner" %in% queries_accdb) {
    testthat::skip(
      "Expected query 'qryCostsSummedByOwner' not present in sample accdb."
    )
  }
  query_sql <- mdb_queries(
    sample_accdb,
    query = "qryCostsSummedByOwner",
    as_list = FALSE
  )
  expect_type(query_sql, "character")
  expect_true(grepl("^SELECT", query_sql, ignore.case = TRUE))
})

test_that("test_sql.sh is covered by mdb_sql input mode", {
  skip_if_not(file.exists(sample_mdb))
  skip_if_not(file.exists(sample_sql))

  sql_text <- mdb_sql(
    path = sample_mdb,
    input = sample_sql,
    as_text = TRUE,
    no_pretty_print = TRUE,
    no_footer = TRUE
  )

  expect_type(sql_text, "character")
  expect_true(nchar(sql_text) > 0)
})

test_that("test_sql script is replicated in DBI context", {
  skip_if_not(file.exists(sample_mdb))
  skip_if_not(file.exists(sample_sql))

  conn <- DBI::dbConnect(sample_mdb)
  on.exit(DBI::dbDisconnect(conn), add = TRUE)

  statements <- read_sql_statements(sample_sql)
  expect_true(length(statements) >= 1)

  out1 <- DBI::dbGetQuery(conn, statements[[1]])
  expect_s3_class(out1, "data.frame")
  expect_lte(nrow(out1), 10)
  expect_true("CustomerID" %in% names(out1))

  out2 <- DBI::dbGetQuery(conn, statements[[2]])
  expect_s3_class(out2, "data.frame")
  expect_true("City" %in% names(out2))
  expect_true(all(out2$City == "Helsinki"))

  out3 <- DBI::dbGetQuery(conn, statements[[3]])
  expect_s3_class(out3, "data.frame")
  expect_true("CompanyName" %in% names(out3))
  expect_gt(nrow(out3), 0)
})

library(testthat)
library(mdbr)

test_that("schema returns a col_spec with expected columns", {
  skip_on_cran()
  skip_if_not(is.loaded("mdbr_version"))
  skip_if_not_installed("readr")
  dat <- mdb_schema(mdb_example(), "Flights")
  expect_s3_class(dat, "col_spec")
  expect_length(dat$cols, 19L)
})

test_that("schema condense returns condensed col_spec", {
  skip_on_cran()
  skip_if_not(is.loaded("mdbr_version"))
  skip_if_not_installed("readr")
  dat <- mdb_schema(mdb_example(), "Flights", condense = TRUE)
  expect_s3_class(dat, "col_spec")
})

test_that("schema errors without table", {
  skip_on_cran()
  skip_if_not(is.loaded("mdbr_version"))
  expect_error(mdb_schema(mdb_example()))
})

library(testthat)
library(mdbr)

# Tests that verify the bundled mdbtools functionality using the nycflights13
# example database shipped with the mdbr package.

test_that("mdb_tables returns tables using native backend", {
  skip_if_not(is.loaded("mdbr_version"))
  t <- mdb_tables(mdb_example())
  expect_type(t, "character")
  expect_gte(length(t), 1L)
  expect_true("Airlines" %in% t)
})

test_that("mdb_tables type='query' returns character vector", {
  skip_if_not(is.loaded("mdbr_version"))
  q <- mdb_tables(mdb_example(), type = "query")
  expect_type(q, "character")
})

test_that("mdb_tables returns a character vector", {
  skip_if_not(is.loaded("mdbr_version"))
  txt <- mdb_tables(mdb_example())
  expect_type(txt, "character")
  expect_gte(length(txt), 1L)
})

test_that("mdb_ver returns version string from native library", {
  skip_if_not(is.loaded("mdbr_version"))
  ver <- mdb_ver()
  expect_type(ver, "character")
  expect_length(ver, 1L)
  expect_true(nzchar(ver))
})

test_that("mdb_ver with path returns file format", {
  skip_if_not(is.loaded("mdbr_version"))
  fmt <- mdb_ver(mdb_example())
  expect_type(fmt, "character")
  expect_true(grepl("^(JET|ACE)", fmt))
})

test_that("mdb_sql queries nycflights13 example", {
  skip_if_not(is.loaded("mdbr_version"))
  df <- mdb_sql(mdb_example(), "SELECT * FROM [Airlines] LIMIT 3;")
  expect_s3_class(df, "data.frame")
  expect_lte(nrow(df), 3L)
  expect_true("carrier" %in% names(df) || ncol(df) >= 1)
})

test_that("mdb_count returns integer row count", {
  skip_if_not(is.loaded("mdbr_version"))
  n <- mdb_count(mdb_example(), "Airlines")
  expect_type(n, "integer")
  expect_gte(n, 1L)
})

test_that("mdb_schema returns DDL text for a table", {
  skip_if_not(is.loaded("mdbr_version"))
  ddl <- mdb_schema(
    mdb_example(),
    table = "Airlines",
    mode = "ddl",
    as_list = FALSE
  )
  expect_type(ddl, "character")
  expect_true(grepl("CREATE TABLE", ddl, ignore.case = TRUE))
})

test_that("mdb_export returns CSV text", {
  skip_if_not(is.loaded("mdbr_version"))
  csv <- mdb_export(mdb_example(), "Airlines", n = 2L)
  expect_type(csv, "character")
  expect_length(csv, 1L)
  expect_true(grepl(",", csv))
})

test_that("DBI connection to nycflights13 example works", {
  skip_if_not(is.loaded("mdbr_version"))
  conn <- DBI::dbConnect(mdb(), dbname = mdb_example())
  on.exit(DBI::dbDisconnect(conn), add = TRUE)

  expect_true(DBI::dbIsValid(conn))
  tables <- DBI::dbListTables(conn)
  expect_true("Airlines" %in% tables)

  df <- DBI::dbReadTable(conn, "Airlines")
  expect_s3_class(df, "data.frame")
  expect_gte(nrow(df), 1L)
})

test_that("export_mdb backward-compat: returns CSV text via native", {
  skip_if_not(is.loaded("mdbr_version"))
  dat <- export_mdb(mdb_example(), "Airlines", output = TRUE)
  expect_type(dat, "character")
  expect_length(dat, 1L)
})

test_that("read_mdb backward-compat: reads table as tibble", {
  skip_if_not(is.loaded("mdbr_version"))
  dat <- read_mdb(mdb_example(), "Airlines")
  expect_s3_class(dat, "data.frame")
  expect_gte(nrow(dat), 1L)
})

#' Export an Access database table as a text file
#'
#' Convert the data of a table into a delimited text string. Save the string as
#' a character vector or write it to a text file. This direct conversion makes
#' it easy to read tables into R or a spreadsheet.
#'
#' @param file Path to the Microsoft Access file.
#' @param table Name of the table, list with `mdb_tables()`.
#' @param output Controls where output is sent. `TRUE` (the default) returns
#'   the output as a character vector. `""` prints to the R console and returns
#'   invisibly. `NULL` or `FALSE` discards the output. A character string is
#'   treated as a file path to write to, returning the path invisibly.
#' @param delim Delimiter used to separate values.
#' @param quote Single character used to quote strings. Defaults to `"`.
#' @param quote_escape The type of escaping to use for quoted values, one of
#'   `"double"`, `"backslash"` or `"none"`. You can also use `FALSE`, which is
#'   equivalent to "none". The default is `"double"`, which is expected format
#'   for Excel.
#' @param col_names If `FALSE`, column names will not be included at the top of
#'   the file. If `TRUE`, column names will be included.
#' @param eol The end of line character to use. Most commonly either `"\n"` for
#'   Unix style newlines, or `"\r\n"` for Windows style newlines.
#' @param date_format The format in which date columns are converted. MDB Tools
#'   uses the `strftime(3)` format, similar to [readr::parse_date()]. No need to
#'   specify whole string. Defaults to ISO8601.
#' @return Character string, invisible if path to file.
#' @examples
#' \dontrun{
#' export_mdb(mdb_example(), "Airlines", output = TRUE)
#' }
#' @export
export_mdb <- function(
  file,
  table,
  output = TRUE,
  delim = ",",
  quote = "\"",
  quote_escape = "double",
  col_names = TRUE,
  eol = "\n",
  date_format = "%Y-%m-%d %H:%M:%S"
) {
  if (missing(table)) {
    stop("Must define a table name, list with mdb_tables()", call. = FALSE)
  }
  if (identical(output, FALSE) || is.null(output)) {
    return(invisible(NULL))
  }
  quote_escape <- switch(
    standardize_escape(quote_escape),
    double = "\"",
    backslash = "\\",
    none = ""
  )
  escape_arg <- if (nzchar(quote_escape)) quote_escape else NULL

  out <- mdb_export(
    path = file,
    table = table,
    no_header = !col_names,
    delimiter = delim,
    row_delimiter = eol,
    quote = quote,
    escape = escape_arg,
    datetime_format = date_format
  )

  if (isTRUE(output)) {
    return(out)
  } else if (identical(output, "")) {
    cat(out)
    return(invisible(out))
  } else {
    con <- file(output, open = "wb")
    on.exit(close(con), add = TRUE)
    writeLines(out, con = con, useBytes = TRUE)
    return(invisible(output))
  }
}

# from tidyverse/readr/R/write.R
standardize_escape <- function(x) {
  if (identical(x, FALSE)) {
    x <- "none"
  }

  escape_types <- c("double" = 1L, "backslash" = 2L, "none" = 3L)
  escape <- match.arg(tolower(x), names(escape_types))

  escape_types[escape]
}

library(testthat)
library(mdbr)

test_that("tables can be exported as strings", {
  skip_on_cran()
  skip_if_not(is.loaded("mdbr_version"))
  dat <- export_mdb(mdb_example(), "Airlines", output = TRUE)
  expect_type(dat, "character")
  expect_length(dat, 1)
})

test_that("tables can be exported to file", {
  skip_on_cran()
  skip_if_not(is.loaded("mdbr_version"))
  tmp <- tempfile()
  dat <- export_mdb(mdb_example(), "Airlines", output = tmp)
  expect_true(file.exists(tmp))
  expect_equal(file.size(tmp), 450)
})

test_that("tables can be exported without escape", {
  skip_on_cran()
  skip_if_not(is.loaded("mdbr_version"))
  dat <- export_mdb(
    file = mdb_example(),
    table = "Airlines",
    output = TRUE,
    quote_escape = FALSE
  )
  expect_type(dat, "character")
  expect_length(dat, 1)
})

test_that("exporting errors without table name", {
  skip_on_cran()
  skip_if_not(is.loaded("mdbr_version"))
  expect_error(export_mdb(mdb_example()))
})

#' Create an mdbr DBI Driver
#'
#' `mdb()` is the canonical DBI-style constructor for connecting to Microsoft
#' Access '.mdb' and '.accdb' files.
#'
#' @return A DBI driver for '.mdb' and '.accdb' files.
#' @examples
#' db <- mdb_example()
#' conn <- DBI::dbConnect(mdb(), dbname = db)
#' DBI::dbListTables(conn)
#' DBI::dbDisconnect(conn)
#' @useDynLib mdbr
#' @importFrom methods new setClass setMethod
#' @importFrom DBI dbConnect dbDisconnect dbIsValid dbListTables dbListObjects dbExistsTable dbListFields dbReadTable dbQuoteIdentifier dbSendQuery dbGetQuery dbExecute dbFetch dbHasCompleted dbClearResult SQL dbCanConnect
#' @export
mdb <- function() {
  methods::new("MdbDriver")
}

methods::setClass("MdbDriver", contains = "DBIDriver")

methods::setClass(
  "MdbConnection",
  contains = "DBIConnection",
  slots = c(
    path = "character",
    open = "logical"
  )
)

methods::setClass(
  "MdbResult",
  contains = "DBIResult",
  slots = c(
    data = "data.frame",
    position = "integer",
    completed = "logical"
  )
)

methods::setMethod(
  "dbCanConnect",
  "MdbDriver",
  function(drv, ...) {
    tryCatch(
      {
        conn <- DBI::dbConnect(drv, ...)
        on.exit(DBI::dbDisconnect(conn), add = TRUE)
        TRUE
      },
      error = function(e) {
        FALSE
      }
    )
  }
)

methods::setMethod(
  "dbConnect",
  "MdbDriver",
  function(drv, dbname, ...) {
    if (missing(dbname) || !is.character(dbname) || length(dbname) != 1L) {
      stop("`dbname` must be a single '.mdb' or '.accdb' path.", call. = FALSE)
    }

    path <- normalizePath(dbname, mustWork = TRUE)
    methods::new("MdbConnection", path = path, open = TRUE)
  }
)

methods::setMethod(
  "dbConnect",
  "character",
  function(drv, ...) {
    if (.is_mdb_path(drv)) {
      return(DBI::dbConnect(mdb(), dbname = drv, ...))
    }

    stop(
      "When using a character first argument, provide an '.mdb' or '.accdb' path, ",
      "or call DBI::dbConnect(mdb(), dbname = ...).",
      call. = FALSE
    )
  }
)

methods::setMethod(
  "dbDisconnect",
  "MdbConnection",
  function(conn, ...) {
    conn@open <- FALSE
    TRUE
  }
)

methods::setMethod(
  "dbIsValid",
  "MdbConnection",
  function(dbObj, ...) {
    isTRUE(dbObj@open) && file.exists(dbObj@path)
  }
)

methods::setMethod(
  "dbListTables",
  "MdbConnection",
  function(conn, ...) {
    .require_valid_connection(conn)
    .native_list_tables(conn@path)
  }
)

methods::setMethod(
  "dbListObjects",
  "MdbConnection",
  function(conn, prefix = NULL, ...) {
    .require_valid_connection(conn)

    if (!is.null(prefix)) {
      return(data.frame(
        table = I(list()),
        is_prefix = logical(0),
        stringsAsFactors = FALSE
      ))
    }

    tables <- DBI::dbListTables(conn)
    queries <- .native_list_queries(conn@path)
    object_names <- c(tables, queries)
    object_types <- c(
      rep("table", length(tables)),
      rep("query", length(queries))
    )

    data.frame(
      table = I(as.list(object_names)),
      is_prefix = rep(FALSE, length(object_names)),
      .type = object_types,
      stringsAsFactors = FALSE
    )
  }
)

methods::setMethod(
  "dbExistsTable",
  c("MdbConnection", "character"),
  function(conn, name, ...) {
    .require_valid_connection(conn)
    .as_table_name(name) %in% DBI::dbListTables(conn)
  }
)

methods::setMethod(
  "dbExistsTable",
  c("MdbConnection", "Id"),
  function(conn, name, ...) {
    DBI::dbExistsTable(conn, .as_table_name(name))
  }
)

methods::setMethod(
  "dbListFields",
  c("MdbConnection", "character"),
  function(conn, name, ...) {
    .require_valid_connection(conn)
    table_name <- .as_table_name(name)
    .native_list_fields(conn@path, table_name)
  }
)

methods::setMethod(
  "dbListFields",
  c("MdbConnection", "Id"),
  function(conn, name, ...) {
    DBI::dbListFields(conn, .as_table_name(name))
  }
)

methods::setMethod(
  "dbReadTable",
  c("MdbConnection", "character"),
  function(conn, name, ...) {
    .require_valid_connection(conn)
    table_name <- .as_table_name(name)
    native <- .native_read_table(conn@path, table_name)
    .coerce_mdb_data_frame(.as_data_frame(native), native)
  }
)

methods::setMethod(
  "dbReadTable",
  c("MdbConnection", "Id"),
  function(conn, name, ...) {
    DBI::dbReadTable(conn, .as_table_name(name), ...)
  }
)

methods::setMethod(
  "dbQuoteIdentifier",
  c("MdbConnection", "character"),
  function(conn, x, ...) {
    DBI::SQL(paste0("[", gsub("]", "]]", x, fixed = TRUE), "]"))
  }
)

methods::setMethod(
  "dbQuoteIdentifier",
  c("MdbConnection", "Id"),
  function(conn, x, ...) {
    DBI::dbQuoteIdentifier(conn, as.character(unlist(x, use.names = FALSE)))
  }
)

methods::setMethod(
  "dbSendQuery",
  c("MdbConnection", "character"),
  function(conn, statement, ...) {
    .require_valid_connection(conn)
    native <- .native_run_query(conn@path, statement)
    data <- .coerce_mdb_data_frame(.as_data_frame(native), native)
    methods::new(
      "MdbResult",
      data = data,
      position = 0L,
      completed = nrow(data) == 0L
    )
  }
)

methods::setMethod(
  "dbGetQuery",
  c("MdbConnection", "character"),
  function(conn, statement, ...) {
    res <- DBI::dbSendQuery(conn, statement, ...)
    on.exit(DBI::dbClearResult(res), add = TRUE)
    DBI::dbFetch(res, n = -1)
  }
)

methods::setMethod(
  "dbExecute",
  c("MdbConnection", "character"),
  function(conn, statement, ...) {
    .require_valid_connection(conn)
    stop(
      "`dbExecute()` is not supported for MDB/ACCDB in read-only mode.",
      call. = FALSE
    )
  }
)

methods::setMethod(
  "dbFetch",
  "MdbResult",
  function(res, n = -1, ...) {
    start <- res@position + 1L
    total <- nrow(res@data)

    if (n < 0 || is.infinite(n)) {
      end <- total
    } else {
      end <- min(total, res@position + as.integer(n))
    }

    if (start > total) {
      res@completed <- TRUE
      return(res@data[0, , drop = FALSE])
    }

    chunk <- res@data[start:end, , drop = FALSE]
    res@position <- as.integer(end)
    res@completed <- end >= total
    chunk
  }
)

methods::setMethod(
  "dbHasCompleted",
  "MdbResult",
  function(res, ...) {
    isTRUE(res@completed)
  }
)

methods::setMethod(
  "dbIsValid",
  "MdbResult",
  function(dbObj, ...) {
    isTRUE(!dbObj@completed || dbObj@position <= nrow(dbObj@data))
  }
)

methods::setMethod(
  "dbClearResult",
  "MdbResult",
  function(res, ...) {
    res@completed <- TRUE
    TRUE
  }
)

# types from mdbtools/src/libmdb/backend.c
#   MDB Tools - A library for reading MS Access database files
#   Copyright (C) 2000-2011 Brian Bruns and others
mdb_col_types <- list(
  "Unknown 0x00" = "c",
  "Boolean" = "l",
  "Byte" = "i",
  "Integer" = "i",
  "Long Integer" = "i",
  "Currency" = "d",
  "Single" = "d",
  "Double" = "d",
  "DateTime" = "T",
  "Binary" = "c",
  "Text" = "c",
  "OLE" = "c",
  "Memo/Hyperlink" = "c",
  "Unknown 0x0d" = "c",
  "Unknown 0x0e" = "c",
  "Replication ID" = "c",
  "Numeric" = "d"
)

list_switch <- function(val, list) {
  do.call("switch", c(val, as.list(list)))
}

#' Get path to mdbr example
#'
#' mdbr comes bundled with a sample file from the
#' [nycflights13](https://github.com/tidyverse/nycflights13) package in its
#' inst/extdata directory. This function make it easy to access.
#'
#' @param path path to the Microsoft Access file.
#' @return A character string with the full path to the bundled example file.
#' @examples
#' mdb_example()
#' @export
mdb_example <- function(path = "nycflights13.mdb") {
  if (!is.character(path)) {
    dir(system.file("extdata", package = "mdbr"))
  } else {
    system.file("extdata", path, package = "mdbr", mustWork = TRUE)
  }
}

library(testthat)
library(mdbr)

test_that("tables can be listed as vector", {
  skip_on_cran()
  skip_if_not(is.loaded("mdbr_version"))
  t <- mdb_tables(mdb_example())
  expect_type(t, "character")
  expect_length(t, 4)
})

#' List tables in a Microsoft Access database
#'
#' `mdb-tables` is a utility program distributed with MDB Tools.
#' It outputs the names of all user tables (or other object types) in an MDB
#' database file.
#'
#' @param file Path to the Microsoft Access file.
#' @param system Logical; include system (`MSys*`) tables. Equivalent to `-S`.
#' @param type Object type to list: `"table"` (default), `"query"`,
#'   `"systable"`, `"any"`, `"all"`, `"form"`, `"macro"`, `"report"`,
#'   `"linkedtable"`, `"module"`, `"relationship"`, or `"dbprop"`.
#'   Equivalent to `-t`.
#' @param show_type Logical; prefix each entry with its type. Equivalent to `-T`.
#' @return A character vector of object names.
#' @examples
#' db <- mdb_example()
#' mdb_tables(db)
#' mdb_tables(db, type = "query")
#' @export
mdb_tables <- function(
  file,
  system = FALSE,
  type = c(
    "table",
    "query",
    "systable",
    "any",
    "all",
    "form",
    "macro",
    "report",
    "linkedtable",
    "module",
    "relationship",
    "dbprop"
  ),
  show_type = FALSE
) {
  type <- match.arg(type)
  path <- .mdb_normalize_path(file)

  # Map R type names to C integer codes passed to mdbr_list_objects:
  #  1   = user tables (MDB_TABLE + !system flag)
  # -3   = system tables (MDB_TABLE + system flag)
  #  5   = queries    (MDB_QUERY)
  #  0   = forms      (MDB_FORM)
  #  2   = macros     (MDB_MACRO)
  #  4   = reports    (MDB_REPORT)
  #  6   = linked tables (MDB_LINKED_TABLE)
  #  7   = modules    (MDB_MODULE)
  #  8   = relationships (MDB_RELATIONSHIP)
  # 10   = db properties (MDB_DATABASE_PROPERTY)
  # -2   = all non-system
  # -1   = all (MDB_ANY)
  type_codes <- c(
    table = 1L,
    systable = -3L,
    query = 5L,
    form = 0L,
    macro = 2L,
    report = 4L,
    linkedtable = 6L,
    module = 7L,
    relationship = 8L,
    dbprop = 10L,
    any = -2L,
    all = -1L
  )

  # For any/all with show_type, make per-type calls so we can label each entry.
  if (isTRUE(show_type) && type %in% c("any", "all")) {
    single_types <- setdiff(names(type_codes), c("any", "all"))
    if (type == "any") {
      single_types <- setdiff(single_types, "systable")
    }
    parts <- character(0)
    for (t in single_types) {
      entries <- .native_list_objects(path, type_codes[[t]])
      if (length(entries)) parts <- c(parts, paste(t, entries))
    }
    if (isTRUE(system) && type == "any") {
      sys_entries <- .native_list_objects(path, -3L)
      if (length(sys_entries)) parts <- c(parts, paste("systable", sys_entries))
    }
    return(parts)
  }

  # Single call for all other cases.
  out <- .native_list_objects(path, type_codes[[type]])

  # system = TRUE appends system tables on top of type result
  # (except systable/all which already contain them)
  if (isTRUE(system) && !type %in% c("systable", "all")) {
    sys_out <- .native_list_objects(path, -3L)
    if (isTRUE(show_type)) {
      out <- paste(type, out)
      sys_out <- paste("systable", sys_out)
    }
    return(c(out, sys_out))
  }

  if (isTRUE(show_type)) {
    out <- paste(type, out)
  }

  out
}

# Internal native wrappers and helpers for the bundled mdbtools library.
# These functions call into the compiled C library (src/mdb_native.c).

.is_mdb_path <- function(x) {
  is.character(x) &&
    length(x) == 1L &&
    grepl("\\.(mdb|accdb)$", x, ignore.case = TRUE)
}

.as_table_name <- function(name) {
  if (inherits(name, "Id")) {
    vals <- unlist(name, use.names = FALSE)
    return(as.character(vals[[length(vals)]]))
  }
  as.character(name[[1]])
}

.require_valid_connection <- function(conn) {
  if (!DBI::dbIsValid(conn)) {
    stop("Invalid or closed MDB connection.", call. = FALSE)
  }
}

.native_list_tables <- function(path, system = FALSE) {
  .Call("mdbr_list_tables", PACKAGE = "mdbr", path, system)
}

.native_list_queries <- function(path) {
  .Call("mdbr_list_queries", PACKAGE = "mdbr", path)
}

.native_list_objects <- function(path, type_int) {
  .Call("mdbr_list_objects", PACKAGE = "mdbr", path, as.integer(type_int))
}

.native_list_fields <- function(path, table) {
  .Call("mdbr_list_fields", PACKAGE = "mdbr", path, table)
}

.native_table_num_rows <- function(path, table) {
  .Call("mdbr_table_num_rows", PACKAGE = "mdbr", path, table)
}

.native_read_table <- function(path, table) {
  .Call("mdbr_read_table", PACKAGE = "mdbr", path, table)
}

.native_run_query <- function(path, statement) {
  .Call("mdbr_run_query", PACKAGE = "mdbr", path, statement)
}

.native_get_query_sql <- function(path, query_name) {
  .Call("mdbr_get_query_sql", PACKAGE = "mdbr", path, query_name)
}

.native_print_schema <- function(
  path,
  table = NULL,
  backend = NULL,
  namespace = NULL,
  export_options = NULL
) {
  .Call(
    "mdbr_print_schema",
    PACKAGE = "mdbr",
    path,
    table,
    backend,
    namespace,
    export_options
  )
}

.native_mdbtools_version <- function() {
  .Call("mdbr_version", PACKAGE = "mdbr")
}

.native_file_format <- function(path) {
  .Call("mdbr_file_format", PACKAGE = "mdbr", path)
}

.native_prop_dump <- function(path, name, propcol = NULL) {
  .Call("mdbr_prop_dump", PACKAGE = "mdbr", path, name, propcol)
}

.trim_sql_semicolon <- function(x) {
  x <- trimws(x)
  sub(";\\s*$", "", x)
}

.expand_saved_query_statement <- function(path, statement) {
  pattern <- "^\\s*SELECT\\s+\\*\\s+FROM\\s+\\[([^\\]]+)\\]\\s*(?:LIMIT\\s+([0-9]+))?\\s*;?\\s*$"
  m <- regexec(pattern, statement, ignore.case = TRUE, perl = TRUE)
  hit <- regmatches(statement, m)[[1]]
  if (length(hit) == 0L) {
    return(statement)
  }

  query_name <- hit[[2]]
  limit <- if (length(hit) >= 3L) hit[[3]] else ""
  queries <- .native_list_queries(path)
  if (!query_name %in% queries) {
    return(statement)
  }

  query_sql <- .native_get_query_sql(path, query_name)
  query_sql <- .trim_sql_semicolon(query_sql)
  if (nzchar(limit)) {
    query_sql <- paste0(query_sql, " LIMIT ", limit)
  }
  query_sql
}

.as_data_frame <- function(x) {
  if (length(x) == 0L) {
    return(data.frame())
  }
  as.data.frame(
    x,
    check.names = FALSE,
    stringsAsFactors = FALSE,
    optional = TRUE
  )
}

.MDB_TYPE <- list(
  BOOL = 0x01L,
  BYTE = 0x02L,
  INT = 0x03L,
  LONGINT = 0x04L,
  MONEY = 0x05L,
  FLOAT = 0x06L,
  DOUBLE = 0x07L,
  DATETIME = 0x08L,
  BINARY = 0x09L,
  TEXT = 0x0aL,
  OLE = 0x0bL,
  MEMO = 0x0cL,
  REPID = 0x0fL,
  NUMERIC = 0x10L,
  COMPLEX = 0x12L
)

.normalize_string_na <- function(x) {
  x <- trimws(as.character(x))
  x[x == ""] <- NA_character_
  x
}

.coerce_logical <- function(x) {
  out <- rep(NA, length(x))
  if (!length(x)) {
    return(out)
  }
  y <- tolower(trimws(as.character(x)))
  false_vals <- c("0", "false", "f", "no", "n")
  true_vals <- c("1", "-1", "true", "t", "yes", "y")
  out[y %in% false_vals] <- FALSE
  out[y %in% true_vals] <- TRUE
  out[y == ""] <- NA
  out
}

.coerce_datetime <- function(x) {
  y <- .normalize_string_na(x)
  out <- rep(as.POSIXct(NA_real_, origin = "1970-01-01", tz = "UTC"), length(y))
  if (!length(y)) {
    return(out)
  }
  formats <- c(
    "%Y-%m-%d %H:%M:%OS",
    "%Y-%m-%d %H:%M",
    "%Y-%m-%d",
    "%m/%d/%Y %H:%M:%OS",
    "%m/%d/%Y %H:%M",
    "%m/%d/%Y",
    "%d/%m/%Y %H:%M:%OS",
    "%d/%m/%Y %H:%M",
    "%d/%m/%Y"
  )
  pending <- !is.na(y)
  for (fmt in formats) {
    if (!any(pending)) {
      break
    }
    parsed <- as.POSIXct(y[pending], format = fmt, tz = "UTC")
    ok <- !is.na(parsed)
    if (any(ok)) {
      idx <- which(pending)[ok]
      out[idx] <- parsed[ok]
      pending[idx] <- FALSE
    }
  }
  out
}

.coerce_column_by_type <- function(x, type_code) {
  type_code <- as.integer(type_code[[1]])
  if (is.na(type_code) || !length(x)) {
    return(x)
  }
  if (type_code == .MDB_TYPE$BOOL) {
    return(.coerce_logical(x))
  }
  if (type_code %in% c(.MDB_TYPE$BYTE, .MDB_TYPE$INT, .MDB_TYPE$LONGINT)) {
    return(suppressWarnings(as.integer(.normalize_string_na(x))))
  }
  if (
    type_code %in%
      c(.MDB_TYPE$MONEY, .MDB_TYPE$FLOAT, .MDB_TYPE$DOUBLE, .MDB_TYPE$NUMERIC)
  ) {
    return(suppressWarnings(as.numeric(.normalize_string_na(x))))
  }
  if (type_code == .MDB_TYPE$DATETIME) {
    return(.coerce_datetime(x))
  }
  x
}

.coerce_mdb_data_frame <- function(df, source) {
  type_codes <- attr(source, "mdb_col_types", exact = TRUE)
  if (is.null(type_codes) || !length(type_codes) || !ncol(df)) {
    return(df)
  }
  n <- min(length(type_codes), ncol(df))
  for (i in seq_len(n)) {
    df[[i]] <- .coerce_column_by_type(df[[i]], type_codes[[i]])
  }
  df
}

.mdb_normalize_path <- function(path) {
  normalizePath(path, mustWork = TRUE)
}

.mdb_connect <- function(path) {
  DBI::dbConnect(mdb(), dbname = .mdb_normalize_path(path))
}

.mdb_schema_options <- function(
  drop_table = FALSE,
  not_null = TRUE,
  default_values = FALSE,
  not_empty = FALSE,
  comments = TRUE,
  indexes = TRUE,
  relations = TRUE
) {
  opts <- 0L
  if (isTRUE(drop_table)) {
    opts <- bitwOr(opts, 1L)
  }
  if (isTRUE(not_null)) {
    opts <- bitwOr(opts, bitwShiftL(1L, 1L))
  }
  if (isTRUE(not_empty)) {
    opts <- bitwOr(opts, bitwShiftL(1L, 2L))
  }
  if (isTRUE(comments)) {
    opts <- bitwOr(opts, bitwShiftL(1L, 3L))
  }
  if (isTRUE(default_values)) {
    opts <- bitwOr(opts, bitwShiftL(1L, 4L))
  }
  if (isTRUE(indexes)) {
    opts <- bitwOr(opts, bitwShiftL(1L, 5L))
  }
  if (isTRUE(relations)) {
    opts <- bitwOr(opts, bitwShiftL(1L, 6L))
  }
  as.integer(opts)
}

.mdb_quote_ident <- function(x) {
  x <- as.character(x)
  x <- gsub("]", "]]", x, fixed = TRUE)
  paste0("[", x, "]")
}

.mdb_field_text <- function(
  x,
  null = "",
  no_quote = FALSE,
  quote = '"',
  escape = NULL,
  escape_invisible = FALSE
) {
  if (is.na(x)) {
    return(null)
  }
  value <- as.character(x)
  if (escape_invisible) {
    value <- gsub("\\\\", "\\\\\\\\", value, fixed = TRUE, useBytes = TRUE)
    value <- gsub("\\r", "\\\\r", value, fixed = TRUE, useBytes = TRUE)
    value <- gsub("\\n", "\\\\n", value, fixed = TRUE, useBytes = TRUE)
    value <- gsub("\\t", "\\\\t", value, fixed = TRUE, useBytes = TRUE)
  }
  if (isTRUE(no_quote)) {
    return(value)
  }
  if (is.null(escape)) {
    value <- gsub(
      quote,
      paste0(quote, quote),
      value,
      fixed = TRUE,
      useBytes = TRUE
    )
  } else {
    value <- gsub(
      quote,
      paste0(escape, quote),
      value,
      fixed = TRUE,
      useBytes = TRUE
    )
  }
  paste0(quote, value, quote)
}

.mdb_apply_datetime_formats <- function(
  df,
  date_format = "%Y-%m-%d",
  datetime_format = "%Y-%m-%d %H:%M:%S"
) {
  out <- df
  for (nm in names(out)) {
    if (inherits(out[[nm]], "POSIXct")) {
      out[[nm]] <- ifelse(
        is.na(out[[nm]]),
        NA_character_,
        format(out[[nm]], datetime_format, tz = "UTC")
      )
    } else if (inherits(out[[nm]], "Date")) {
      out[[nm]] <- ifelse(
        is.na(out[[nm]]),
        NA_character_,
        format(out[[nm]], date_format)
      )
    }
  }
  out
}

.mdb_apply_boolean_words <- function(df, boolean_words = FALSE) {
  out <- df
  for (nm in names(out)) {
    if (is.logical(out[[nm]])) {
      if (isTRUE(boolean_words)) {
        out[[nm]] <- ifelse(
          is.na(out[[nm]]),
          NA_character_,
          ifelse(out[[nm]], "TRUE", "FALSE")
        )
      } else {
        out[[nm]] <- ifelse(
          is.na(out[[nm]]),
          NA_integer_,
          ifelse(out[[nm]], 1L, 0L)
        )
      }
    }
  }
  out
}

.mdb_apply_unprintable <- function(df, no_unprintable = FALSE) {
  if (!isTRUE(no_unprintable)) {
    return(df)
  }
  out <- df
  for (nm in names(out)) {
    if (is.character(out[[nm]])) {
      out[[nm]] <- gsub("[^[:print:]\\t\\r\\n]", " ", out[[nm]], perl = TRUE)
    }
  }
  out
}

.mdb_delimited_text <- function(
  df,
  delimiter = "\t",
  row_delimiter = "\n",
  header = TRUE,
  null = "",
  no_quote = FALSE,
  quote = '"',
  escape = NULL,
  escape_invisible = FALSE
) {
  rows <- character(0)
  if (isTRUE(header)) {
    rows <- c(rows, paste(names(df), collapse = delimiter))
  }
  if (nrow(df) > 0L) {
    body <- apply(df, 1L, function(row) {
      vals <- vapply(
        row,
        .mdb_field_text,
        FUN.VALUE = character(1),
        null = null,
        no_quote = no_quote,
        quote = quote,
        escape = escape,
        escape_invisible = escape_invisible
      )
      paste(vals, collapse = delimiter)
    })
    rows <- c(rows, unname(body))
  }
  paste(rows, collapse = row_delimiter)
}

.mdb_sql_literal <- function(x) {
  if (is.na(x)) {
    return("NULL")
  }
  if (is.numeric(x)) {
    return(as.character(x))
  }
  if (is.logical(x)) {
    return(if (isTRUE(x)) "1" else "0")
  }
  val <- as.character(x)
  val <- gsub("'", "''", val, fixed = TRUE)
  paste0("'", val, "'")
}

.mdb_query_table <- function(path, table, n = -1L) {
  table <- as.character(table[[1]])
  sql <- sprintf("SELECT * FROM %s", .mdb_quote_ident(table))
  if (!is.null(n) && is.finite(n) && n >= 0) {
    sql <- paste(sql, "LIMIT", as.integer(n))
  }
  mdb_sql(path = path, statement = sql)
}

.mdb_read_sql_input <- function(input) {
  lines <- readLines(input, warn = FALSE)
  lines <- trimws(lines)
  lines <- lines[nzchar(lines)]
  lines <- lines[!startsWith(lines, "#")]
  lines <- lines[tolower(lines) != "go"]
  script <- paste(lines, collapse = "\n")
  parts <- strsplit(script, ";", fixed = TRUE)[[1]]
  parts <- trimws(parts)
  parts <- parts[nzchar(parts)]
  parts
}

.mdb_example_nwind_path <- function() {
  candidates <- c(
    Sys.getenv("MDBR_EXAMPLE_DB", unset = ""),
    system.file(
      "testthat",
      "mdbtestdata",
      "data",
      "nwind.mdb",
      package = "mdbr"
    ),
    system.file(
      "tests",
      "testthat",
      "mdbtestdata",
      "data",
      "nwind.mdb",
      package = "mdbr"
    )
  )
  candidates <- unique(candidates[nzchar(candidates)])
  hits <- candidates[file.exists(candidates)]
  if (!length(hits)) {
    return("")
  }
  normalizePath(hits[[1]], mustWork = TRUE)
}

.as_mdblist <- function(x) {
  if (length(x) == 0L) {
    return(structure(x, class = "mdblist"))
  }
  if (is.null(names(x)) || any(!nzchar(names(x)))) {
    stop("`mdblist` entries must be named.", call. = FALSE)
  }
  structure(x, class = "mdblist")
}

#' Print Method For `mdblist`
#'
#' Pretty printer for multi-object text outputs returned by selected
#' `mdb_*` helpers when `as_list = TRUE` (default).
#'
#' @param x A `mdblist` object.
#' @param ... Unused.
#'
#' @return The input object, invisibly.
#' @export
print.mdblist <- function(x, ...) {
  n <- length(x)
  if (n == 0L) {
    cat("<mdblist[0]>\n")
    return(invisible(x))
  }
  for (i in seq_len(n)) {
    nm <- names(x)[[i]]
    cat("[", nm, "]\n", sep = "")
    val <- x[[i]]
    if (!is.null(val) && length(val) > 0L) {
      cat(as.character(val), sep = "\n")
    }
  }
  invisible(x)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/example.R
\name{mdb_example}
\alias{mdb_example}
\title{Get path to mdbr example}
\usage{
mdb_example(path = "nycflights13.mdb")
}
\arguments{
\item{path}{path to the Microsoft Access file.}
}
\value{
A character string with the full path to the bundled example file.
}
\description{
mdbr comes bundled with a sample file from the
\href{https://github.com/tidyverse/nycflights13}{nycflights13} package in its
inst/extdata directory. This function make it easy to access.
}
\examples{
mdb_example()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/advanced.R
\name{mdb_prop}
\alias{mdb_prop}
\title{Get Properties List From MDB Database}
\usage{
mdb_prop(
  path,
  name = NULL,
  propcol = "LvProp",
  version = FALSE,
  as_list = TRUE
)
}
\arguments{
\item{path}{Path to \code{.mdb}/\code{.accdb} file.}

\item{name}{Object name (\code{table}, \code{query}, etc.); may be a character vector.}

\item{propcol}{Property column name. Defaults to \code{LvProp}.}

\item{version}{Logical; when \code{TRUE}, return mdbtools version.}

\item{as_list}{Logical; defaults to \code{TRUE}. When \code{TRUE}, wraps the result in
an \code{mdblist} object so the dedicated print method is used.}
}
\value{
A named list with one entry per element of \code{name}. Each entry is
itself a named list of named character vectors, one per property block
(e.g. \code{"(none)"} for table-level properties, or a column/field name).
Access individual values with \code{p[["Orders"]][["(none)"]]["Description"]}.
}
\description{
\code{mdb-prop} retrieves properties for one or more objects in an MDB database.
\code{name} is the name of the table, query, or other object.
\code{propcol} is the name of the \code{MSysObjects} column containing properties and
defaults to \code{LvProp}.
}
\examples{
db <- mdbr:::.mdb_example_nwind_path()
if (nzchar(db)) {
  p <- mdb_prop(db, "Orders")
  p[["Orders"]][["(none)"]]["Description"]
  p2 <- mdb_prop(db, c("Orders", "Orders Qry"))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/advanced.R
\name{mdb_export}
\alias{mdb_export}
\title{Export Data In An MDB Table To CSV Or INSERT SQL}
\usage{
mdb_export(
  path,
  table,
  no_header = FALSE,
  delimiter = ",",
  row_delimiter = "\\n",
  no_quote = FALSE,
  quote = "\\"",
  escape = NULL,
  escape_invisible = FALSE,
  date_format = "\%Y-\%m-\%d",
  datetime_format = "\%Y-\%m-\%d \%H:\%M:\%S",
  null = "",
  bin = c("strip", "raw", "octal", "hex"),
  boolean_words = FALSE,
  insert = NULL,
  namespace = NULL,
  batch_size = 1L,
  n = -1L
)
}
\arguments{
\item{path}{Path to \code{.mdb}/\code{.accdb} file.}

\item{table}{Table name.}

\item{no_header}{Logical, equivalent to \code{-H/--no-header}.}

\item{delimiter}{Equivalent to \code{-d/--delimiter}.}

\item{row_delimiter}{Equivalent to \code{-R/--row-delimiter}.}

\item{no_quote}{Equivalent to \code{-Q/--no-quote}.}

\item{quote}{Equivalent to \code{-q/--quote}.}

\item{escape}{Equivalent to \code{-X/--escape}.}

\item{escape_invisible}{Equivalent to \code{-e/--escape-invisible}.}

\item{date_format}{Equivalent to \code{-D/--date-format}.}

\item{datetime_format}{Equivalent to \code{-T/--datetime-format}.}

\item{null}{Equivalent to \code{-0/--null}.}

\item{bin}{Binary mode (\code{strip}, \code{raw}, \code{octal}, \code{hex}) for parity.}

\item{boolean_words}{Equivalent to \code{-B/--boolean-words}.}

\item{insert}{Backend for \code{-I/--insert} mode.}

\item{namespace}{Equivalent to \code{-N/--namespace}.}

\item{batch_size}{Equivalent to \code{-S/--batch-size}.}

\item{n}{Optional row limit (\verb{LIMIT n}).}
}
\value{
Character scalar containing CSV or SQL INSERT text.
}
\description{
\code{mdb-export} is a utility program distributed with MDB Tools.
It produces CSV output for the given table. Such output is suitable for
importation into databases or spreadsheets.
}
\details{
Used with \code{insert}, it outputs backend-specific SQL \code{INSERT} statements.
Most formatting options also apply in insert mode.
}
\examples{
db <- mdbr:::.mdb_example_nwind_path()
if (nzchar(db)) {
  cat(mdb_export(db, "Products", n = 2))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/advanced.R
\name{mdb_count}
\alias{mdb_count}
\title{Count Rows In Table}
\usage{
mdb_count(path, table = NULL, where = NULL, version = FALSE)
}
\arguments{
\item{path}{Path to \code{.mdb}/\code{.accdb} file.}

\item{table}{Table name.}

\item{where}{Optional SQL predicate appended to \code{WHERE}. This is an R-side
extension and is not part of the CLI.}

\item{version}{Logical; when \code{TRUE}, return mdbtools version (\code{--version}).}
}
\value{
Integer row count or version string.
}
\description{
\code{mdb-count} is a utility program distributed with MDB Tools.
It outputs the number of rows in a table.
}
\examples{
db <- mdbr:::.mdb_example_nwind_path()
if (nzchar(db)) {
  mdb_count(db, "Orders")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tables.R
\name{mdb_tables}
\alias{mdb_tables}
\title{List tables in a Microsoft Access database}
\usage{
mdb_tables(
  file,
  system = FALSE,
  type = c("table", "query", "systable", "any", "all", "form", "macro", "report",
    "linkedtable", "module", "relationship", "dbprop"),
  show_type = FALSE
)
}
\arguments{
\item{file}{Path to the Microsoft Access file.}

\item{system}{Logical; include system (\verb{MSys*}) tables. Equivalent to \code{-S}.}

\item{type}{Object type to list: \code{"table"} (default), \code{"query"},
\code{"systable"}, \code{"any"}, \code{"all"}, \code{"form"}, \code{"macro"}, \code{"report"},
\code{"linkedtable"}, \code{"module"}, \code{"relationship"}, or \code{"dbprop"}.
Equivalent to \code{-t}.}

\item{show_type}{Logical; prefix each entry with its type. Equivalent to \code{-T}.}
}
\value{
A character vector of object names.
}
\description{
\code{mdb-tables} is a utility program distributed with MDB Tools.
It outputs the names of all user tables (or other object types) in an MDB
database file.
}
\examples{
db <- mdb_example()
mdb_tables(db)
mdb_tables(db, type = "query")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/advanced.R
\name{mdb_array}
\alias{mdb_array}
\title{Export Table As List Columns (mdb-array mimic)}
\usage{
mdb_array(path, table, columns = NULL, n = -1L)
}
\arguments{
\item{path}{Path to \code{.mdb}/\code{.accdb} file.}

\item{table}{Table name.}

\item{columns}{Optional character vector of columns.}

\item{n}{Optional row limit (\verb{LIMIT n}).}
}
\value{
Named list, one entry per selected column.
}
\description{
\code{mdb-array(1)} emits C source; this wrapper returns a named R list of columns
while keeping equivalent database/table inputs.
}
\examples{
db <- mdbr:::.mdb_example_nwind_path()
if (nzchar(db)) {
  arr <- mdb_array(db, "Products", columns = c("ProductID", "ProductName"), n = 2)
  str(arr)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/advanced.R
\name{mdb_hexdump}
\alias{mdb_hexdump}
\title{Hexdump MDB File (mdb-hexdump mimic)}
\usage{
mdb_hexdump(path, pagenumber = NULL, page_size = 4096L, n = 256L)
}
\arguments{
\item{path}{Path to file.}

\item{pagenumber}{Optional page index (0-based) like \verb{mdb-hexdump file [pagenumber]}.}

\item{page_size}{Page size in bytes (default 4096 for modern Jet/ACE).}

\item{n}{Number of bytes to emit.}
}
\value{
Single hexadecimal string.
}
\description{
Hexdump MDB File (mdb-hexdump mimic)
}
\examples{
db <- mdbr:::.mdb_example_nwind_path()
if (nzchar(db)) {
  mdb_hexdump(db, n = 16)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/advanced.R
\name{mdb_queries}
\alias{mdb_queries}
\title{List Or Retrieve Queries In An MDB Database}
\usage{
mdb_queries(
  path,
  query = NULL,
  list = TRUE,
  newline = FALSE,
  delimiter = " ",
  as_text = FALSE,
  as_list = TRUE
)
}
\arguments{
\item{path}{Path to \code{.mdb}/\code{.accdb} file.}

\item{query}{Character vector of query names. When \code{NULL} (the default),
the function lists available query names instead of fetching SQL.}

\item{list}{Logical; when \code{TRUE} (the default) and \code{query} is \code{NULL},
return the list of query names. Set to \code{FALSE} to suppress listing and
return \code{character(0)}.}

\item{newline}{Logical; when \code{TRUE} and \code{as_text = TRUE}, collapse names
with \code{"\\n"} instead of \code{delimiter}.}

\item{delimiter}{Character scalar used to collapse query names when
\code{as_text = TRUE}. Default \code{" "}.}

\item{as_text}{Logical; when \code{TRUE}, collapses the query name list into a
single string using \code{delimiter} (or \code{"\\n"} if \code{newline = TRUE}).}

\item{as_list}{Logical; defaults to \code{TRUE}. When \code{TRUE} and \code{query} is
supplied, wraps the result in a named \code{mdblist} object.}
}
\value{
When \code{query} is \code{NULL}: a character vector of query names (or a
collapsed string when \code{as_text = TRUE}). When \code{query} is supplied and
\code{as_list = TRUE}: a named \code{mdblist} of SQL text strings, one per query.
With a single query and \code{as_list = FALSE}: a character scalar.
}
\description{
\code{mdb-queries} is a utility program distributed with MDB Tools.
Without \code{query}, it lists the names of all saved queries in the database.
With \code{query}, it returns the SQL text of the named query or queries.
}
\examples{
db <- mdbr:::.mdb_example_nwind_path()
if (nzchar(db)) {
  mdb_queries(db)
  mdb_queries(db, "Orders Qry")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/advanced.R
\name{mdb_import}
\alias{mdb_import}
\title{Import CSV Into MDB (mdb-import mimic)}
\usage{
mdb_import(path, table, csvfile, header_lines = 0L, delimiter = ",")
}
\arguments{
\item{path}{Path to \code{.mdb}/\code{.accdb} file.}

\item{table}{Table name.}

\item{csvfile}{CSV file path.}

\item{header_lines}{Equivalent to \code{-H/--header} skipped lines.}

\item{delimiter}{Equivalent to \code{-d/--delimiter}.}
}
\value{
No return; always errors in read-only mode.
}
\description{
\code{mdb-import(1)} writes to MDB files. This package is read-only, so this
function validates CLI-like options and then errors.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/export.R
\name{export_mdb}
\alias{export_mdb}
\title{Export an Access database table as a text file}
\usage{
export_mdb(
  file,
  table,
  output = TRUE,
  delim = ",",
  quote = "\\"",
  quote_escape = "double",
  col_names = TRUE,
  eol = "\\n",
  date_format = "\%Y-\%m-\%d \%H:\%M:\%S"
)
}
\arguments{
\item{file}{Path to the Microsoft Access file.}

\item{table}{Name of the table, list with \code{mdb_tables()}.}

\item{output}{Controls where output is sent. \code{TRUE} (the default) returns
the output as a character vector. \code{""} prints to the R console and returns
invisibly. \code{NULL} or \code{FALSE} discards the output. A character string is
treated as a file path to write to, returning the path invisibly.}

\item{delim}{Delimiter used to separate values.}

\item{quote}{Single character used to quote strings. Defaults to \verb{"}.}

\item{quote_escape}{The type of escaping to use for quoted values, one of
\code{"double"}, \code{"backslash"} or \code{"none"}. You can also use \code{FALSE}, which is
equivalent to "none". The default is \code{"double"}, which is expected format
for Excel.}

\item{col_names}{If \code{FALSE}, column names will not be included at the top of
the file. If \code{TRUE}, column names will be included.}

\item{eol}{The end of line character to use. Most commonly either \code{"\\n"} for
Unix style newlines, or \code{"\\r\\n"} for Windows style newlines.}

\item{date_format}{The format in which date columns are converted. MDB Tools
uses the \code{strftime(3)} format, similar to \code{\link[readr:parse_date]{readr::parse_date()}}. No need to
specify whole string. Defaults to ISO8601.}
}
\value{
Character string, invisible if path to file.
}
\description{
Convert the data of a table into a delimited text string. Save the string as
a character vector or write it to a text file. This direct conversion makes
it easy to read tables into R or a spreadsheet.
}
\examples{
\dontrun{
export_mdb(mdb_example(), "Airlines", output = TRUE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/advanced.R
\name{mdb_json}
\alias{mdb_json}
\title{Export Data In An MDB Table To JSON}
\usage{
mdb_json(
  path,
  table,
  date_format = "\%Y-\%m-\%d",
  time_format = "\%Y-\%m-\%d \%H:\%M:\%S",
  no_unprintable = FALSE,
  n = -1L
)
}
\arguments{
\item{path}{Path to \code{.mdb}/\code{.accdb} file.}

\item{table}{Table name.}

\item{date_format}{Equivalent to \code{-D/--date-format}.}

\item{time_format}{Equivalent to \code{-T/--time-format}.}

\item{no_unprintable}{Equivalent to \code{-U/--no-unprintable}.}

\item{n}{Optional row limit.}
}
\value{
JSON string.
}
\description{
\code{mdb-json} is a utility program distributed with MDB Tools.
It produces JSON output for the given table. Such output is suitable for
parsing in a variety of languages.
}
\examples{
db <- mdbr:::.mdb_example_nwind_path()
if (nzchar(db) && requireNamespace("jsonlite", quietly = TRUE)) {
  mdb_json(db, "Products", n = 2)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/advanced.R
\name{mdb_sql}
\alias{mdb_sql}
\title{SQL Interface To MDB Tools}
\usage{
mdb_sql(
  path,
  statement = NULL,
  no_header = FALSE,
  no_footer = FALSE,
  no_pretty_print = FALSE,
  delimiter = "\\t",
  input = NULL,
  output = NULL,
  as_text = FALSE
)
}
\arguments{
\item{path}{Path to \code{.mdb}/\code{.accdb} file.}

\item{statement}{SQL statement text.}

\item{no_header}{Logical, equivalent to \code{-H/--no-header} (for text mode).}

\item{no_footer}{Logical, equivalent to \code{-F/--no-footer} (for text mode).}

\item{no_pretty_print}{Logical, equivalent to \code{-p/--no-pretty-print}.}

\item{delimiter}{Delimiter equivalent to \code{-d/--delimiter} in plain text mode.}

\item{input}{Input file equivalent to \code{-i/--input}.}

\item{output}{Output file equivalent to \code{-o/--output}.}

\item{as_text}{Logical; when \code{TRUE}, returns CLI-like text output.}
}
\value{
\code{data.frame} by default, or character scalar in text mode.
}
\description{
\code{mdb-sql} is a utility program distributed with MDB Tools.
It allows querying of an MDB database using a limited SQL subset language.
The supported SQL is intentionally small: single-table queries, no aggregates,
and limited \code{WHERE} support.
}
\details{
In addition to single statements, this wrapper accepts \code{input} files similar
to \verb{mdb-sql -i file}, strips \code{go} batch terminators, and executes the script
one statement at a time.
}
\examples{
db <- mdbr:::.mdb_example_nwind_path()
if (nzchar(db)) {
  mdb_sql(db, "SELECT [ProductID], [ProductName] FROM [Products] LIMIT 3;")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/native.R
\name{print.mdblist}
\alias{print.mdblist}
\title{Print Method For \code{mdblist}}
\usage{
\method{print}{mdblist}(x, ...)
}
\arguments{
\item{x}{A \code{mdblist} object.}

\item{...}{Unused.}
}
\value{
The input object, invisibly.
}
\description{
Pretty printer for multi-object text outputs returned by selected
\verb{mdb_*} helpers when \code{as_list = TRUE} (default).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dbi.R
\name{mdb}
\alias{mdb}
\title{Create an mdbr DBI Driver}
\usage{
mdb()
}
\value{
A DBI driver for '.mdb' and '.accdb' files.
}
\description{
\code{mdb()} is the canonical DBI-style constructor for connecting to Microsoft
Access '.mdb' and '.accdb' files.
}
\examples{
db <- mdb_example()
conn <- DBI::dbConnect(mdb(), dbname = db)
DBI::dbListTables(conn)
DBI::dbDisconnect(conn)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/advanced.R
\name{mdb_header}
\alias{mdb_header}
\title{MDB Header Summary (mdb-header mimic)}
\usage{
mdb_header(path)
}
\arguments{
\item{path}{Path to \code{.mdb}/\code{.accdb} file.}
}
\value{
Named list with version, table names and query names.
}
\description{
\code{mdb-header(1)} writes C files; this wrapper returns a structured summary.
}
\examples{
db <- mdbr:::.mdb_example_nwind_path()
if (nzchar(db)) {
  mdb_header(db)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/advanced.R
\name{mdb_ver}
\alias{mdb_ver}
\title{Return MDB File Format Or MDB Tools Version}
\usage{
mdb_ver(path = NULL, version = FALSE)
}
\arguments{
\item{path}{Optional database path. When omitted, the wrapper returns the
mdbtools package version for backward compatibility.}

\item{version}{Logical, equivalent to \code{-M/--version}.}
}
\value{
Single character string with file format or mdbtools version.
}
\description{
\code{mdb-ver} will return a single line of output corresponding to the program
that produced the file: \code{JET3} (for files produced by Access 97), \code{JET4}
(Access 2000, XP and 2003), \code{ACE12} (Access 2007), \code{ACE14} (Access 2010),
\code{ACE15} (Access 2013), or \code{ACE16} (Access 2016).
}
\examples{
mdb_ver()
db <- mdbr:::.mdb_example_nwind_path()
if (nzchar(db)) {
  mdb_ver(db)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/advanced.R
\name{mdb_schema}
\alias{mdb_schema}
\title{Generate Schema DDL or Column Types}
\usage{
mdb_schema(
  path,
  table = NULL,
  mode = c("legacy", "ddl"),
  condense = FALSE,
  namespace = NULL,
  backend = c("access", "sybase", "oracle", "postgres", "mysql", "sqlite"),
  drop_table = FALSE,
  not_null = TRUE,
  default_values = FALSE,
  not_empty = FALSE,
  comments = TRUE,
  indexes = TRUE,
  relations = TRUE,
  as_list = TRUE
)
}
\arguments{
\item{path}{Path to \code{.mdb}/\code{.accdb} file.}

\item{table}{Table name(s). For \code{mode = "ddl"}, defaults to all user tables.
For \code{mode = "legacy"}, exactly one table name must be given.}

\item{mode}{\code{"legacy"} (default) returns a \code{\link[readr:cols]{readr::cols()}} specification,
matching the k5cents/mdbr canonical API. Requires the \pkg{readr} package.
\code{"ddl"} returns DDL text.}

\item{condense}{Logical; only used when \code{mode = "legacy"}. When \code{TRUE},
passes the col spec through \code{\link[readr:cols_condense]{readr::cols_condense()}}. Default \code{FALSE}.}

\item{namespace}{Prefix identifiers with namespace, equivalent to
\code{-N/--namespace}. Only used when \code{mode = "ddl"}.}

\item{backend}{Target DDL dialect. Supported values are \code{access}, \code{sybase},
\code{oracle}, \code{postgres}, \code{mysql}, and \code{sqlite}. Only used when \code{mode = "ddl"}.}

\item{drop_table}{Issue \verb{DROP TABLE} statements.}

\item{not_null}{Include \verb{NOT NULL} constraints.}

\item{default_values}{Include \code{DEFAULT} values.}

\item{not_empty}{Include \verb{CHECK <> ''} constraints.}

\item{comments}{Include \verb{COMMENT ON} statements.}

\item{indexes}{Export indexes.}

\item{relations}{Request foreign key constraints. Current library-mode
implementation emits a placeholder comment; full FK export is not yet
implemented.}

\item{as_list}{Logical; defaults to \code{TRUE}. When \code{TRUE}, returns a named
\code{mdblist} of DDL text entries keyed by table name. Only used when
\code{mode = "ddl"}.}
}
\value{
When \code{mode = "legacy"}, a \code{\link[readr:cols]{readr::cols()}} specification (optionally
condensed via \code{\link[readr:cols_condense]{readr::cols_condense()}}). Requires the \pkg{readr} package.
When \code{mode = "ddl"} and \code{as_list = TRUE}, a named \code{mdblist} of table-level
DDL text.
}
\description{
\code{mdb-schema} is a utility program distributed with MDB Tools. It produces
DDL (data definition language) output for the given database, which can be
passed to another database to recreate the Access table structure. With
\code{mode = "legacy"} (the default), it returns a
\link[readr:cols]{readr col spec} for the table instead.
}
\examples{
db <- mdbr:::.mdb_example_nwind_path()
if (nzchar(db)) {
  mdb_schema(db, table = "Products")
  mdb_schema(db, table = "Products", mode = "ddl")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read.R
\name{read_mdb}
\alias{read_mdb}
\title{Read a table as data frame}
\usage{
read_mdb(file, table, col_names = TRUE, col_types = NULL, ...)
}
\arguments{
\item{file}{Path to the Microsoft Access file.}

\item{table}{Name of the table, list with \code{\link[=mdb_tables]{mdb_tables()}}.}

\item{col_names}{Logical; when \code{FALSE} columns are named \code{V1}, \code{V2}, etc.}

\item{col_types}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Ignored. Type coercion
is now handled automatically by the native reader.}

\item{...}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Ignored. Extra arguments were
previously forwarded to \code{\link[readr:read_delim]{readr::read_delim()}}, which is no longer used.}
}
\value{
A \link[tibble:tibble]{tibble}.
}
\description{
Reads a table directly from a Microsoft Access database using the bundled
mdbtools C library. Column types are inferred from the MDB schema:
integer, double, logical, \link[base:DateTimeClasses]{POSIXct} for DateTime
columns, and character otherwise.
}
\examples{
\dontrun{
read_mdb(mdb_example(), "Airlines")
}
}
